import os
import sys
import shutil
import json
import time
import datetime
import tempfile
import warnings
from collections import defaultdict
from typing import Optional

import tensorflow as tf
from tensorflow.python import pywrap_tensorflow
from tensorflow.core.util import event_pb2
from tensorflow.python.util import compat

from neorl.rl.baselines.shared.misc_util import mpi_rank_or_zero

DEBUG = 10
INFO = 20
WARN = 30
ERROR = 40

DISABLED = 50


class KVWriter(object):
    """
    Key Value writer
    """
    def writekvs(self, kvs):
        """
        write a dictionary to file

        :param kvs: (dict)
        """
        raise NotImplementedError


class SeqWriter(object):
    """
    sequence writer
    """
    def writeseq(self, seq):
        """
        write an array to file

        :param seq: (list)
        """
        raise NotImplementedError


class HumanOutputFormat(KVWriter, SeqWriter):
    def __init__(self, filename_or_file):
        """
        log to a file, in a human readable format

        :param filename_or_file: (str or File) the file to write the log to
        """
        if isinstance(filename_or_file, str):
            self.file = open(filename_or_file, 'wt')
            self.own_file = True
        else:
            assert hasattr(filename_or_file, 'write'), 'Expected file or str, got {}'.format(filename_or_file)
            self.file = filename_or_file
            self.own_file = False

    def writekvs(self, kvs):
        # Create strings for printing
        key2str = {}
        for (key, val) in sorted(kvs.items()):
            if isinstance(val, float):
                valstr = '%-8.3g' % (val,)
            else:
                valstr = str(val)
            key2str[self._truncate(key)] = self._truncate(valstr)

        # Find max widths
        if len(key2str) == 0:
            warnings.warn('Tried to write empty key-value dict')
            return
        else:
            keywidth = max(map(len, key2str.keys()))
            valwidth = max(map(len, key2str.values()))

        # Write out the data
        dashes = '-' * (keywidth + valwidth + 7)
        lines = [dashes]
        for (key, val) in sorted(key2str.items()):
            lines.append('| %s%s | %s%s |' % (
                key,
                ' ' * (keywidth - len(key)),
                val,
                ' ' * (valwidth - len(val)),
            ))
        lines.append(dashes)
        self.file.write('\n'.join(lines) + '\n')

        # Flush the output to the file
        self.file.flush()

    @classmethod
    def _truncate(cls, string):
        return string[:20] + '...' if len(string) > 23 else string

    def writeseq(self, seq):
        seq = list(seq)
        for (i, elem) in enumerate(seq):
            self.file.write(elem)
            if i < len(seq) - 1:  # add space unless this is the last one
                self.file.write(' ')
        self.file.write('\n')
        self.file.flush()

    def close(self):
        """
        closes the file
        """
        if self.own_file:
            self.file.close()


class JSONOutputFormat(KVWriter):
    def __init__(self, filename):
        """
        log to a file, in the JSON format

        :param filename: (str) the file to write the log to
        """
        self.file = open(filename, 'wt')

    def writekvs(self, kvs):
        for key, value in sorted(kvs.items()):
            if hasattr(value, 'dtype'):
                if value.shape == () or len(value) == 1:
                    # if value is a dimensionless numpy array or of length 1, serialize as a float
                    kvs[key] = float(value)
                else:
                    # otherwise, a value is a numpy array, serialize as a list or nested lists
                    kvs[key] = value.tolist()
        self.file.write(json.dumps(kvs) + '\n')
        self.file.flush()

    def close(self):
        """
        closes the file
        """
        self.file.close()


class CSVOutputFormat(KVWriter):
    def __init__(self, filename):
        """
        log to a file, in a CSV format

        :param filename: (str) the file to write the log to
        """
        self.file = open(filename, 'w+t')
        self.keys = []
        self.sep = ','

    def writekvs(self, kvs):
        # Add our current row to the history
        extra_keys = kvs.keys() - self.keys
        if extra_keys:
            self.keys.extend(extra_keys)
            self.file.seek(0)
            lines = self.file.readlines()
            self.file.seek(0)
            for (i, key) in enumerate(self.keys):
                if i > 0:
                    self.file.write(',')
                self.file.write(key)
            self.file.write('\n')
            for line in lines[1:]:
                self.file.write(line[:-1])
                self.file.write(self.sep * len(extra_keys))
                self.file.write('\n')
        for i, key in enumerate(self.keys):
            if i > 0:
                self.file.write(',')
            value = kvs.get(key)
            if value is not None:
                self.file.write(str(value))
        self.file.write('\n')
        self.file.flush()

    def close(self):
        """
        closes the file
        """
        self.file.close()


def summary_val(key, value):
    """
    :param key: (str)
    :param value: (float)
    """
    kwargs = {'tag': key, 'simple_value': float(value)}
    return tf.Summary.Value(**kwargs)


def valid_float_value(value):
    """
    Returns True if the value can be successfully cast into a float

    :param value: (Any) the value to check
    :return: (bool)
    """
    try:
        float(value)
        return True
    except TypeError:
        return False


class TensorBoardOutputFormat(KVWriter):
    def __init__(self, folder):
        """
        Dumps key/value pairs into TensorBoard's numeric format.

        :param folder: (str) the folder to write the log to
        """
        os.makedirs(folder, exist_ok=True)
        self.dir = folder
        self.step = 1
        prefix = 'events'
        path = os.path.join(os.path.abspath(folder), prefix)
        self.writer = pywrap_tensorflow.EventsWriter(compat.as_bytes(path))  # type: pywrap_tensorflow.EventsWriter

    def writekvs(self, kvs):
        summary = tf.Summary(value=[summary_val(k, v) for k, v in kvs.items() if valid_float_value(v)])
        event = event_pb2.Event(wall_time=time.time(), summary=summary)
        event.step = self.step  # is there any reason why you'd want to specify the step?
        if self.writer is None:
            raise ValueError("Attempt to write after close().")
        self.writer.WriteEvent(event)
        self.writer.Flush()
        self.step += 1

    def close(self):
        """
        closes the file
        """
        if self.writer:
            self.writer.Close()
            self.writer = None


def make_output_format(_format, ev_dir, log_suffix=''):
    """
    return a logger for the requested format

    :param _format: (str) the requested format to log to ('stdout', 'log', 'json', 'csv' or 'tensorboard')
    :param ev_dir: (str) the logging directory
    :param log_suffix: (str) the suffix for the log file
    :return: (KVWrite) the logger
    """
    os.makedirs(ev_dir, exist_ok=True)
    if _format == 'stdout':
        return HumanOutputFormat(sys.stdout)
    elif _format == 'log':
        return HumanOutputFormat(os.path.join(ev_dir, 'log%s.txt' % log_suffix))
    elif _format == 'json':
        return JSONOutputFormat(os.path.join(ev_dir, 'progress%s.json' % log_suffix))
    elif _format == 'csv':
        return CSVOutputFormat(os.path.join(ev_dir, 'progress%s.csv' % log_suffix))
    elif _format == 'tensorboard':
        return TensorBoardOutputFormat(os.path.join(ev_dir, 'tb%s' % log_suffix))
    else:
        raise ValueError('Unknown format specified: %s' % (_format,))


# ================================================================
# API
# ================================================================

def logkv(key, val):
    """
    Log a value of some diagnostic
    Call this once for each diagnostic quantity, each iteration
    If called many times, last value will be used.

    :param key: (Any) save to log this key
    :param val: (Any) save to log this value
    """
    Logger.CURRENT.logkv(key, val)


def logkv_mean(key, val):
    """
    The same as logkv(), but if called many times, values averaged.

    :param key: (Any) save to log this key
    :param val: (Number) save to log this value
    """
    Logger.CURRENT.logkv_mean(key, val)


def logkvs(key_values):
    """
    Log a dictionary of key-value pairs

    :param key_values: (dict) the list of keys and values to save to log
    """
    for key, value in key_values.items():
        logkv(key, value)


def dumpkvs():
    """
    Write all of the diagnostics from the current iteration
    """
    Logger.CURRENT.dumpkvs()


def getkvs():
    """
    get the key values logs

    :return: (dict) the logged values
    """
    return Logger.CURRENT.name2val


def log(*args, level=INFO):
    """
    Write the sequence of args, with no separators,
    to the console and output files (if you've configured an output file).

    level: int. (see logger.py docs) If the global logger level is higher than
                the level argument here, don't print to stdout.

    :param args: (list) log the arguments
    :param level: (int) the logging level (can be DEBUG=10, INFO=20, WARN=30, ERROR=40, DISABLED=50)
    """
    Logger.CURRENT.log(*args, level=level)


def debug(*args):
    """
    Write the sequence of args, with no separators,
    to the console and output files (if you've configured an output file).
    Using the DEBUG level.

    :param args: (list) log the arguments
    """
    log(*args, level=DEBUG)


def info(*args):
    """
    Write the sequence of args, with no separators,
    to the console and output files (if you've configured an output file).
    Using the INFO level.

    :param args: (list) log the arguments
    """
    log(*args, level=INFO)


def warn(*args):
    """
    Write the sequence of args, with no separators,
    to the console and output files (if you've configured an output file).
    Using the WARN level.

    :param args: (list) log the arguments
    """
    log(*args, level=WARN)


def error(*args):
    """
    Write the sequence of args, with no separators,
    to the console and output files (if you've configured an output file).
    Using the ERROR level.

    :param args: (list) log the arguments
    """
    log(*args, level=ERROR)


def set_level(level):
    """
    Set logging threshold on current logger.

    :param level: (int) the logging level (can be DEBUG=10, INFO=20, WARN=30, ERROR=40, DISABLED=50)
    """
    Logger.CURRENT.set_level(level)


def get_level():
    """
    Get logging threshold on current logger.
    :return: (int) the logging level (can be DEBUG=10, INFO=20, WARN=30, ERROR=40, DISABLED=50)
    """
    return Logger.CURRENT.level


def get_dir():
    """
    Get directory that log files are being written to.
    will be None if there is no output directory (i.e., if you didn't call start)

    :return: (str) the logging directory
    """
    return Logger.CURRENT.get_dir()


record_tabular = logkv
dump_tabular = dumpkvs


class ProfileKV:
    def __init__(self, name):
        """
        Usage:
        with logger.ProfileKV("interesting_scope"):
            code

        :param name: (str) the profiling name
        """
        self.name = "wait_" + name

    def __enter__(self):
        self.start_time = time.time()

    def __exit__(self, _type, value, traceback):
        Logger.CURRENT.name2val[self.name] += time.time() - self.start_time


def profile(name):
    """
    Usage:
    @profile("my_func")
    def my_func(): code

    :param name: (str) the profiling name
    :return: (function) the wrapped function
    """
    def decorator_with_name(func):
        def func_wrapper(*args, **kwargs):
            with ProfileKV(name):
                return func(*args, **kwargs)

        return func_wrapper

    return decorator_with_name


# ================================================================
# Backend
# ================================================================

class Logger(object):
    # A logger with no output files. (See right below class definition)
    #  So that you can still log to the terminal without setting up any output files
    DEFAULT = None  # type: Optional["Logger"]
    # Current logger being used by the free functions above
    CURRENT = None  # type: Optional["Logger"]

    def __init__(self, folder, output_formats):
        """
        the logger class

        :param folder: (str) the logging location
        :param output_formats: ([str]) the list of output format
        """
        self.name2val = defaultdict(float)  # values this iteration
        self.name2cnt = defaultdict(int)
        self.level = INFO
        self.dir = folder
        self.output_formats = output_formats

    # Logging API, forwarded
    # ----------------------------------------
    def logkv(self, key, val):
        """
        Log a value of some diagnostic
        Call this once for each diagnostic quantity, each iteration
        If called many times, last value will be used.

        :param key: (Any) save to log this key
        :param val: (Any) save to log this value
        """
        self.name2val[key] = val

    def logkv_mean(self, key, val):
        """
        The same as logkv(), but if called many times, values averaged.

        :param key: (Any) save to log this key
        :param val: (Number) save to log this value
        """
        if val is None:
            self.name2val[key] = None
            return
        oldval, cnt = self.name2val[key], self.name2cnt[key]
        self.name2val[key] = oldval * cnt / (cnt + 1) + val / (cnt + 1)
        self.name2cnt[key] = cnt + 1

    def dumpkvs(self):
        """
        Write all of the diagnostics from the current iteration
        """
        if self.level == DISABLED:
            return
        for fmt in self.output_formats:
            if isinstance(fmt, KVWriter):
                fmt.writekvs(self.name2val)
        self.name2val.clear()
        self.name2cnt.clear()

    def log(self, *args, level=INFO):
        """
        Write the sequence of args, with no separators,
        to the console and output files (if you've configured an output file).

        level: int. (see logger.py docs) If the global logger level is higher than
                    the level argument here, don't print to stdout.

        :param args: (list) log the arguments
        :param level: (int) the logging level (can be DEBUG=10, INFO=20, WARN=30, ERROR=40, DISABLED=50)
        """
        if self.level <= level:
            self._do_log(args)

    # Configuration
    # ----------------------------------------
    def set_level(self, level):
        """
        Set logging threshold on current logger.

        :param level: (int) the logging level (can be DEBUG=10, INFO=20, WARN=30, ERROR=40, DISABLED=50)
        """
        self.level = level

    def get_dir(self):
        """
        Get directory that log files are being written to.
        will be None if there is no output directory (i.e., if you didn't call start)

        :return: (str) the logging directory
        """
        return self.dir

    def close(self):
        """
        closes the file
        """
        for fmt in self.output_formats:
            fmt.close()

    # Misc
    # ----------------------------------------
    def _do_log(self, args):
        """
        log to the requested format outputs

        :param args: (list) the arguments to log
        """
        for fmt in self.output_formats:
            if isinstance(fmt, SeqWriter):
                fmt.writeseq(map(str, args))


Logger.DEFAULT = Logger.CURRENT = Logger(folder=None, output_formats=[HumanOutputFormat(sys.stdout)])


def configure(folder=None, format_strs=None):
    """
    configure the current logger

    :param folder: (str) the save location (if None, $OPENAI_LOGDIR, if still None, tempdir/openai-[date & time])
    :param format_strs: (list) the output logging format
        (if None, $OPENAI_LOG_FORMAT, if still None, ['stdout', 'log', 'csv'])
    """
    if folder is None:
        folder = os.getenv('OPENAI_LOGDIR')
    if folder is None:
        folder = os.path.join(tempfile.gettempdir(), datetime.datetime.now().strftime("openai-%Y-%m-%d-%H-%M-%S-%f"))
    assert isinstance(folder, str)
    os.makedirs(folder, exist_ok=True)
    rank = mpi_rank_or_zero()

    log_suffix = ''
    if format_strs is None:
        if rank == 0:
            format_strs = os.getenv('OPENAI_LOG_FORMAT', 'stdout,log,csv').split(',')
        else:
            log_suffix = "-rank%03i" % rank
            format_strs = os.getenv('OPENAI_LOG_FORMAT_MPI', 'log').split(',')
    format_strs = filter(None, format_strs)
    output_formats = [make_output_format(f, folder, log_suffix) for f in format_strs]

    Logger.CURRENT = Logger(folder=folder, output_formats=output_formats)
    log('Logging to %s' % folder)


def reset():
    """
    reset the current logger
    """
    if Logger.CURRENT is not Logger.DEFAULT:
        Logger.CURRENT.close()
        Logger.CURRENT = Logger.DEFAULT
        log('Reset logger')


class ScopedConfigure(object):
    def __init__(self, folder=None, format_strs=None):
        """
        Class for using context manager while logging

        usage:
        with ScopedConfigure(folder=None, format_strs=None):
            {code}

        :param folder: (str) the logging folder
        :param format_strs: ([str]) the list of output logging format
        """
        self.dir = folder
        self.format_strs = format_strs
        self.prevlogger = None

    def __enter__(self):
        self.prevlogger = Logger.CURRENT
        configure(folder=self.dir, format_strs=self.format_strs)

    def __exit__(self, *args):
        Logger.CURRENT.close()
        Logger.CURRENT = self.prevlogger


# ================================================================

def _demo():
    """
    tests for the logger module
    """
    info("hi")
    debug("shouldn't appear")
    set_level(DEBUG)
    debug("should appear")
    folder = "/tmp/testlogging"
    if os.path.exists(folder):
        shutil.rmtree(folder)
    configure(folder=folder)
    logkv("a", 3)
    logkv("b", 2.5)
    dumpkvs()
    logkv("b", -2.5)
    logkv("a", 5.5)
    dumpkvs()
    info("^^^ should see a = 5.5")
    logkv_mean("b", -22.5)
    logkv_mean("b", -44.4)
    logkv("a", 5.5)
    dumpkvs()
    with ScopedConfigure(None, None):
        info("^^^ should see b = 33.3")

    with ScopedConfigure("/tmp/test-logger/", ["json"]):
        logkv("b", -2.5)
        dumpkvs()

    reset()
    logkv("a", "longasslongasslongasslongasslongasslongassvalue")
    dumpkvs()
    warn("hey")
    error("oh")
    logkvs({"test": 1})


# ================================================================
# Readers
# ================================================================

def read_json(fname):
    """
    read a json file using pandas

    :param fname: (str) the file path to read
    :return: (pandas DataFrame) the data in the json
    """
    import pandas
    data = []
    with open(fname, 'rt') as file_handler:
        for line in file_handler:
            data.append(json.loads(line))
    return pandas.DataFrame(data)


def read_csv(fname):
    """
    read a csv file using pandas

    :param fname: (str) the file path to read
    :return: (pandas DataFrame) the data in the csv
    """
    import pandas
    return pandas.read_csv(fname, index_col=None, comment='#')


def read_tb(path):
    """
    read a tensorboard output

    :param path: (str) a tensorboard file OR a directory, where we will find all TB files of the form events.
    :return: (pandas DataFrame) the tensorboad data
    """
    import pandas
    import numpy as np
    from glob import glob
    # from collections import defaultdict
    import tensorflow as tf
    if os.path.isdir(path):
        fnames = glob(os.path.join(path, "events.*"))
    elif os.path.basename(path).startswith("events."):
        fnames = [path]
    else:
        raise NotImplementedError("Expected tensorboard file or directory containing them. Got %s" % path)
    tag2pairs = defaultdict(list)
    maxstep = 0
    for fname in fnames:
        for summary in tf.train.summary_iterator(fname):
            if summary.step > 0:
                for value in summary.summary.value:
                    pair = (summary.step, value.simple_value)
                    tag2pairs[value.tag].append(pair)
                maxstep = max(summary.step, maxstep)
    data = np.empty((maxstep, len(tag2pairs)))
    data[:] = np.nan
    tags = sorted(tag2pairs.keys())
    for (colidx, tag) in enumerate(tags):
        pairs = tag2pairs[tag]
        for (step, value) in pairs:
            data[step - 1, colidx] = value
    return pandas.DataFrame(data, columns=tags)


if __name__ == "__main__":
    _demo()



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Thu Dec  3 14:42:29 2020
#
#@author: Majdi
#"""


import random
import numpy as np
import joblib
from neorl.evolu.discrete import mutate_discrete, encode_grid_to_discrete 
from neorl.evolu.discrete import decode_discrete_to_grid, encode_grid_indv_to_discrete
from neorl.utils.seeding import set_neorl_seed
from neorl.utils.tools import get_population, check_mixed_individual

class GWO(object):
    """
    Grey Wolf Optimizer
    
    :param mode: (str) problem type, either "min" for minimization problem or "max" for maximization
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}``
    :param fit: (function) the fitness function 
    :param nwolves: (int): number of the grey wolves in the group
    :param int_transform: (str): method of handling int/discrete variables, choose from: ``nearest_int``, ``sigmoid``, ``minmax``.
    :param ncores: (int) number of parallel processors (must be ``<= nwolves``)
    :param seed: (int) random seed for sampling
    """
    def __init__(self, mode, bounds, fit, nwolves=5, int_transform ='nearest_int', ncores=1, seed=None):
        
        set_neorl_seed(seed)
        
        assert ncores <= nwolves, '--error: ncores ({}) must be less than or equal than nwolves ({})'.format(ncores, nwolves)
        
        #--mir
        self.mode=mode
        if mode == 'min':
            self.fit=fit
        elif mode == 'max':
            def fitness_wrapper(*args, **kwargs):
                return -fit(*args, **kwargs) 
            self.fit=fitness_wrapper
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')
        
        self.int_transform=int_transform
        if int_transform not in ["nearest_int", "sigmoid", "minmax"]:
            raise ValueError('--error: int_transform entered by user is invalid, must be `nearest_int`, `sigmoid`, or `minmax`')
        
        self.bounds=bounds
        self.ncores = ncores
        self.nwolves=nwolves
        
        #infer variable types 
        self.var_type = np.array([bounds[item][0] for item in bounds])
        
        #mir-grid
        if "grid" in self.var_type:
            self.grid_flag=True
            self.orig_bounds=bounds  #keep original bounds for decoding
            #print('--debug: grid parameter type is found in the space')
            self.bounds, self.bounds_map=encode_grid_to_discrete(self.bounds) #encoding grid to int
            #define var_types again by converting grid to int
            self.var_type = np.array([self.bounds[item][0] for item in self.bounds])
        else:
            self.grid_flag=False
            self.bounds = bounds
            self.orig_bounds=bounds
        
        self.dim = len(bounds)
        self.lb=[self.bounds[item][1] for item in self.bounds]
        self.ub=[self.bounds[item][2] for item in self.bounds]

    def init_sample(self, bounds):
    
        indv=[]
        for key in bounds:
            if bounds[key][0] == 'int':
                indv.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                indv.append(random.uniform(bounds[key][1], bounds[key][2]))
            #elif bounds[key][0] == 'grid':
            #    indv.append(random.sample(bounds[key][1],1)[0])
            else:
                raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')   
        return indv
    
    def ensure_bounds(self, vec): # bounds check

        vec_new = []

        for i, (key, val) in enumerate(self.bounds.items()):
            # less than minimum 
            if vec[i] < self.bounds[key][1]:
                vec_new.append(self.bounds[key][1])
            # more than maximum
            if vec[i] > self.bounds[key][2]:
                vec_new.append(self.bounds[key][2])
            # fine
            if self.bounds[key][1] <= vec[i] <= self.bounds[key][2]:
                vec_new.append(vec[i])
        
        return vec_new
    
    def fit_worker(self, x):
        #This worker is for parallel calculations of the GWO
        
        # Clip the wolf with position outside the lower/upper bounds and return same position
        x=self.ensure_bounds(x)
        
        if self.grid_flag:
            #decode the individual back to the int/float/grid mixed space
            x=decode_discrete_to_grid(x,self.orig_bounds,self.bounds_map)
        
        # Calculate objective function for each search agent
        fitness = self.fit(x)
        
        return fitness

    def ensure_discrete(self, vec):
        #"""
        #to mutate a vector if discrete variables exist 
        #handy function to be used within GWO phases

        #Params:
        #vec - wolf position in vector/list form

        #Return:
        #vec - updated wolf position vector with discrete values
        #"""
        
        for dim in range(self.dim):
            if self.var_type[dim] == 'int':
                vec[dim] = mutate_discrete(x_ij=vec[dim], 
                                               x_min=min(vec),
                                               x_max=max(vec),
                                               lb=self.lb[dim], 
                                               ub=self.ub[dim],
                                               alpha=self.b,
                                               method=self.int_transform,
                                               )
        
        return vec
    
    def evolute(self, ngen, x0=None, verbose=False, **kwargs):
        """
        This function evolutes the GWO algorithm for number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param x0: (list of lists) initial position of the wolves (must be of same size as ``nwolves``)
        :param verbose: (bool) print statistics to screen
        
        :return: (tuple) (best individual, best fitness, and dictionary containing major search results)
        """
        self.history = {'alpha_wolf':[], 'beta_wolf':[], 'delta_wolf': [], 'fitness':[]}
        self.fitness_best=float("inf") 
        self.verbose=verbose
        self.Positions = np.zeros((self.nwolves, self.dim))
        if x0:
            assert len(x0) == self.nwolves, '--error: the length of x0 ({}) MUST equal the number of wolves in the group ({})'.format(len(x0), self.nwolves)
            
            for i in range(self.nwolves):
                check_mixed_individual(x=x0[i], bounds=self.orig_bounds) #assert the type provided is consistent
                if self.grid_flag:
                    self.Positions[i,:] = encode_grid_indv_to_discrete(x0[i], bounds=self.orig_bounds, bounds_map=self.bounds_map)
                else:
                    self.Positions[i,:] = x0[i]
                    
        else:
            # Initialize the positions of search agents
            for i in range(self.nwolves):
                self.Positions[i,:]=self.init_sample(self.bounds)
            
 
        # initialize alpha, beta, and delta_pos
        Alpha_pos = np.zeros(self.dim)
        Alpha_score = float("inf")  #GWO is built to minimize
    
        Beta_pos = np.zeros(self.dim)
        Beta_score = float("inf")  #GWO is built to minimize
    
        Delta_pos = np.zeros(self.dim)
        Delta_score = float("inf") #GWO is built to minimize
           
        with joblib.Parallel(n_jobs=self.ncores) as parallel:
            
            for l in range(0, ngen):
                self.b= 1 - l * ((1) / ngen)  #mir: b decreases linearly between 1 to 0, for discrete mutation
                #---------------------
                # Fitness calcs
                #---------------------
                core_lst=[]
                for case in range (0, self.Positions.shape[0]):
                    core_lst.append(self.Positions[case, :])
            
                if self.ncores > 1:

                    with joblib.Parallel(n_jobs=self.ncores) as parallel:
                        fitness=parallel(joblib.delayed(self.fit_worker)(item) for item in core_lst)
                        
                else:
                    fitness=[]
                    for item in core_lst:
                        fitness.append(self.fit_worker(item))  
                
                self.last_pop=self.Positions.copy()  #for logging
                self.last_fit=np.array(fitness)      #for logging
                #----------------------
                #  Update wolf scores
                #----------------------
                #Loop through the fitness list and update the score of alpha, beta, gamma, and omega!
                for i, fits in enumerate(fitness):
                    # Update Alpha, Beta, and Delta
                    if fits < Alpha_score:
                        Delta_score = Beta_score  # Update delta
                        Delta_pos = Beta_pos.copy()
                        Beta_score = Alpha_score  # Update beta
                        Beta_pos = Alpha_pos.copy()
                        Alpha_score = fits
                        # Update alpha
                        Alpha_pos = self.Positions[i, :].copy()
        
                    if fits > Alpha_score and fits < Beta_score:
                        Delta_score = Beta_score  # Update delte
                        Delta_pos = Beta_pos.copy()
                        Beta_score = fits  # Update beta
                        Beta_pos = self.Positions[i, :].copy()
        
                    if fits > Alpha_score and fits > Beta_score and fits < Delta_score:
                        Delta_score = fits  # Update delta
                        Delta_pos = self.Positions[i, :].copy()
                    
                    #save the best of the best!!!
                    if fits < self.fitness_best:
                        self.fitness_best=fits
                        self.x_best=self.Positions[i, :].copy()
                    
                    
                self.history['alpha_wolf'].append(Alpha_score)
                self.history['beta_wolf'].append(Beta_score)
                self.history['delta_wolf'].append(Delta_score)
                
                if 'a' in kwargs:
                    assert len(kwargs["a"]) == ngen, '--error: the length of `a` in kwargs must equal to ngen'
                    a=kwargs["a"][l]
                else:
                    a = 2 - l * ((2) / ngen)
                # a decreases linearly from 2 to 0
                
                #--------------------------------
                # Position update loop
                #--------------------------------
                # Update the position of search wolves
                for i in range(0, self.nwolves):
                    for j in range(0, self.dim):
        
                        r1 = random.random()  # r1 is a random number in [0,1]
                        r2 = random.random()  # r2 is a random number in [0,1]
        
                        A1 = 2 * a * r1 - a
                        # Equation (3.3)
                        C1 = 2 * r2
                        # Equation (3.4)
                        #print('A1=', A1,C1)
                        D_alpha = abs(C1 * Alpha_pos[j] - self.Positions[i, j])
                        # Equation (3.5)-part 1
                        X1 = Alpha_pos[j] - A1 * D_alpha
                        # Equation (3.6)-part 1
        
                        r1 = random.random()
                        r2 = random.random()
        
                        A2 = 2 * a * r1 - a
                        # Equation (3.3)
                        C2 = 2 * r2
                        # Equation (3.4)
                        #print('A2=', A2,C2)
                        D_beta = abs(C2 * Beta_pos[j] - self.Positions[i, j])
                        # Equation (3.5)-part 2
                        X2 = Beta_pos[j] - A2 * D_beta
                        # Equation (3.6)-part 2
        
                        r1 = random.random()
                        r2 = random.random()
        
                        A3 = 2 * a * r1 - a
                        # Equation (3.3)
                        C3 = 2 * r2
                        # Equation (3.4)
                        #print('A3=', A3,C3)
                        D_delta = abs(C3 * Delta_pos[j] - self.Positions[i, j])
                        # Equation (3.5)-part 3
                        X3 = Delta_pos[j] - A3 * D_delta
                        # Equation (3.5)-part 3
        
                        self.Positions[i, j] = (X1 + X2 + X3) / 3  # Equation (3.7)
                    
                    self.Positions[i,:]=self.ensure_bounds(self.Positions[i,:])
                    self.Positions[i, :] = self.ensure_discrete(self.Positions[i, :])
                 
                #--mir
                if self.mode=='max':
                    self.fitness_best_correct=-self.fitness_best
                else:
                    self.fitness_best_correct=self.fitness_best
                
                # Print statistics
                if self.verbose and i % self.nwolves:
                    print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                    print('GWO step {}/{}, nwolves={}, Ncores={}'.format((l+1)*self.nwolves, ngen*self.nwolves, self.nwolves, self.ncores))
                    print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                    print('Best Group Fitness:', np.round(self.fitness_best_correct,6))
                    if self.grid_flag:
                        self.wolf_decoded = decode_discrete_to_grid(self.x_best, self.orig_bounds, self.bounds_map)
                        print('Best Group Position:', self.wolf_decoded)
                    else:
                        print('Best Group Position:', self.x_best)
                    print('Alpha wolf Fitness:', np.round(Alpha_score,6) if self.mode == 'min' else -np.round(Alpha_score,6))
                    print('Beta wolf Fitness:', np.round(Beta_score,6) if self.mode == 'min' else -np.round(Beta_score,6))
                    print('Delta wolf Fitness:', np.round(Delta_score,6) if self.mode == 'min' else -np.round(Delta_score,6))
                    print('a:', a)
                    print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
    
        
                self.history['fitness'].append(self.fitness_best)

        #mir-grid
        if self.grid_flag:
            self.wolf_correct = decode_discrete_to_grid(self.x_best, self.orig_bounds, self.bounds_map)
        else:
            self.wolf_correct = self.x_best
        
        if self.verbose:
            print('------------------------ GWO Summary --------------------------')
            print('Best fitness (y) found:', self.fitness_best_correct)
            print('Best individual (x) found:', self.wolf_correct)
            print('--------------------------------------------------------------')  
        
        #--mir
        if self.mode=='max':
            self.history['alpha_wolf']=[-item for item in self.history['alpha_wolf']]
            self.history['beta_wolf']=[-item for item in self.history['beta_wolf']]
            self.history['delta_wolf']=[-item for item in self.history['delta_wolf']]
            self.history['fitness']=[-item for item in self.history['fitness']]
            self.last_fit=-self.last_fit
        
        #--mir return the last population for restart calculations
        if self.grid_flag:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=True, 
                                                     bounds=self.orig_bounds, bounds_map=self.bounds_map)
        else:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=False)
            
        return self.wolf_correct, self.fitness_best_correct, self.history




"""This file is used for specifying various schedules that evolve over
time throughout the execution of the algorithm, such as:

 - learning rate for the optimizer
 - exploration epsilon for the epsilon greedy exploration strategy
 - beta parameter for beta parameter in prioritized replay

Each schedule has a function `value(t)` which returns the current value
of the parameter given the timestep t of the optimization procedure.
"""


class Schedule(object):
    def value(self, step):
        """
        Value of the schedule for a given timestep

        :param step: (int) the timestep
        :return: (float) the output value for the given timestep
        """
        raise NotImplementedError


class ConstantSchedule(Schedule):
    """
    Value remains constant over time.

    :param value: (float) Constant value of the schedule
    """

    def __init__(self, value):
        self._value = value

    def value(self, step):
        return self._value


def linear_interpolation(left, right, alpha):
    """
    Linear interpolation between `left` and `right`.

    :param left: (float) left boundary
    :param right: (float) right boundary
    :param alpha: (float) coeff in [0, 1]
    :return: (float)
    """

    return left + alpha * (right - left)


class PiecewiseSchedule(Schedule):
    """
    Piecewise schedule.

    :param endpoints: ([(int, int)])
        list of pairs `(time, value)` meaning that schedule should output
        `value` when `t==time`. All the values for time must be sorted in
        an increasing order. When t is between two times, e.g. `(time_a, value_a)`
        and `(time_b, value_b)`, such that `time_a <= t < time_b` then value outputs
        `interpolation(value_a, value_b, alpha)` where alpha is a fraction of
        time passed between `time_a` and `time_b` for time `t`.
    :param interpolation: (lambda (float, float, float): float)
        a function that takes value to the left and to the right of t according
        to the `endpoints`. Alpha is the fraction of distance from left endpoint to
        right endpoint that t has covered. See linear_interpolation for example.
    :param outside_value: (float)
        if the value is requested outside of all the intervals specified in
        `endpoints` this value is returned. If None then AssertionError is
        raised when outside value is requested.
    """

    def __init__(self, endpoints, interpolation=linear_interpolation, outside_value=None):
        idxes = [e[0] for e in endpoints]
        assert idxes == sorted(idxes)
        self._interpolation = interpolation
        self._outside_value = outside_value
        self._endpoints = endpoints

    def value(self, step):
        for (left_t, left), (right_t, right) in zip(self._endpoints[:-1], self._endpoints[1:]):
            if left_t <= step < right_t:
                alpha = float(step - left_t) / (right_t - left_t)
                return self._interpolation(left, right, alpha)

        # t does not belong to any of the pieces, so doom.
        assert self._outside_value is not None
        return self._outside_value


class LinearSchedule(Schedule):
    """
    Linear interpolation between initial_p and final_p over
    schedule_timesteps. After this many timesteps pass final_p is
    returned.

    :param schedule_timesteps: (int) Number of timesteps for which to linearly anneal initial_p to final_p
    :param initial_p: (float) initial output value
    :param final_p: (float) final output value
    """

    def __init__(self, schedule_timesteps, final_p, initial_p=1.0):
        self.schedule_timesteps = schedule_timesteps
        self.final_p = final_p
        self.initial_p = initial_p

    def value(self, step):
        fraction = min(float(step) / self.schedule_timesteps, 1.0)
        return self.initial_p + fraction * (self.final_p - self.initial_p)


def get_schedule_fn(value_schedule):
    """
    Transform (if needed) learning rate and clip range
    to callable.

    :param value_schedule: (callable or float)
    :return: (function)
    """
    # If the passed schedule is a float
    # create a constant function
    if isinstance(value_schedule, (float, int)):
        # Cast to float to avoid errors
        value_schedule = constfn(float(value_schedule))
    else:
        assert callable(value_schedule)
    return value_schedule


def constfn(val):
    """
    Create a function that returns a constant
    It is useful for learning rate schedule (to avoid code duplication)

    :param val: (float)
    :return: (function)
    """

    def func(_):
        return val

    return func


# ================================================================
# Legacy scheduler used by A2C, AKCTR and ACER
# ================================================================

def constant(_):
    """
    Returns a constant value for the Scheduler

    :param _: ignored
    :return: (float) 1
    """
    return 1.


def linear_schedule(progress):
    """
    Returns a linear value for the Scheduler

    :param progress: (float) Current progress status (in [0, 1])
    :return: (float) 1 - progress
    """
    return 1 - progress


def middle_drop(progress):
    """
    Returns a linear value with a drop near the middle to a constant value for the Scheduler

    :param progress: (float) Current progress status (in [0, 1])
    :return: (float) 1 - progress if (1 - progress) >= 0.75 else 0.075
    """
    eps = 0.75
    if 1 - progress < eps:
        return eps * 0.1
    return 1 - progress


def double_linear_con(progress):
    """
    Returns a linear value (x2) with a flattened tail for the Scheduler

    :param progress: (float) Current progress status (in [0, 1])
    :return: (float) 1 - progress*2 if (1 - progress*2) >= 0.125 else 0.125
    """
    progress *= 2
    eps = 0.125
    if 1 - progress < eps:
        return eps
    return 1 - progress


def double_middle_drop(progress):
    """
    Returns a linear value with two drops near the middle to a constant value for the Scheduler

    :param progress: (float) Current progress status (in [0, 1])
    :return: (float) if 0.75 <= 1 - p: 1 - p, if 0.25 <= 1 - p < 0.75: 0.75, if 1 - p < 0.25: 0.125
    """
    eps1 = 0.75
    eps2 = 0.25
    if 1 - progress < eps1:
        if 1 - progress < eps2:
            return eps2 * 0.5
        return eps1 * 0.1
    return 1 - progress


SCHEDULES = {
    'linear': linear_schedule,
    'constant': constant,
    'double_linear_con': double_linear_con,
    'middle_drop': middle_drop,
    'double_middle_drop': double_middle_drop
}


class Scheduler(object):
    def __init__(self, initial_value, n_values, schedule):
        """
        Update a value every iteration, with a specific curve.

        This is a legacy version of schedules, originally defined
        in a2c/utils.py. Used by A2C, ACER and ACKTR algorithms.

        :param initial_value: (float) initial value
        :param n_values: (int) the total number of iterations
        :param schedule: (function) the curve you wish to follow for your value
        """
        self.step = 0.
        self.initial_value = initial_value
        self.nvalues = n_values
        self.schedule = SCHEDULES[schedule]

    def value(self):
        """
        Update the Scheduler, and return the current value

        :return: (float) the current value
        """
        current_value = self.initial_value * self.schedule(self.step / self.nvalues)
        self.step += 1.
        return current_value

    def value_steps(self, steps):
        """
        Get a value for a given step

        :param steps: (int) The current number of iterations
        :return: (float) the value for the current number of iterations
        """
        return self.initial_value * self.schedule(steps / self.nvalues)



from neorl import PESA
from neorl import PESA2

#Define the fitness function
def FIT(individual):
        y=sum(x**2 for x in individual)
        return y

#Define the fitness function
def FIT2(individual):
        y=sum(x**2 for x in individual)
        return y

#Setup the parameter space (d=5)
nx=5
BOUNDS={}
for i in range(1,nx+1):
        BOUNDS['x'+str(i)]=['float', -100, 100]
            
def test_pesa():

    npop=60
    pesa=PESA(mode='min', bounds=BOUNDS, fit=FIT, npop=npop, mu=40, alpha_init=0.2, 
              alpha_end=1.0, alpha_backdoor=0.1, ncores=9)
    x0=[[50,50,50,50,50] for i in range(npop)]  #initial guess
    x_best, y_best, pesa_hist=pesa.evolute(ngen=4, x0=x0, verbose=1)

test_pesa()

def test_pesa2():
    
    pesa2=PESA2(mode='min', bounds=BOUNDS, fit=FIT2, npop=50, nwolves=5, nwhales=5, ncores=9)
    x_best, y_best, pesa2_hist=pesa2.evolute(ngen=4, replay_every=2, verbose=2)

test_pesa2()


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# ------------------------------------------------------------------------
# This code is taken from the following paper:
#
# Pengfei Huang,Handing Wang,Yaochu Jin,Offine Data-Driven Evolutionary Optimization Based on Tri-Training, Swarm and Evolutionary Computation, Accepted.
#
# You are free to use it for non-commercial purposes. However, we do not offer any forms of guanrantee or warranty associated with the code. We would appreciate your acknowledgement.
# ------------------------------------------------------------------------

import numpy as np
import random
from neorl.hybrid.ngacore.Latin import latin
import warnings
warnings.filterwarnings("ignore")
class GA():
    def __init__(self, pop_size, dimension, lower_bound, upper_bound):
        self.pop_size = pop_size
        self.max_value = upper_bound
        self.min_value = lower_bound
        self.chrom_length = dimension
        self.pc = 1
        self.pm = 1 / dimension
        self.nc = 15
        self.nm = 15
        self.first = []
        self.pop = np.zeros((self.pop_size, self.chrom_length))

    def init_Population(self):
        self.pop = latin(self.pop_size, self.chrom_length, self.min_value, self.max_value)

    def selection(self,fit_value):
        data = np.column_stack((self.pop, fit_value))
        data = data[np.lexsort(data.T)]
        self.pop = data[:self.pop_size, :self.chrom_length]
        self.first.append(self.pop[0])
        return self.pop

    def crossover(self, pc):
        pop_len = len(self.pop)
        for i in range(pop_len):
            if (random.random() < pc):
                populationList = list(range(pop_len))
                populationList.pop(i)
                p2 = random.choice(populationList)
                par1 = self.pop[i]
                par2 = self.pop[p2]
                a = np.row_stack((par1, par2))
                p1 = a.min(axis=0)
                p2 = a.max(axis=0)
                alfa1 = np.array([0.0] * self.chrom_length)
                for ad in range(self.chrom_length):
                    if p1[ad] - self.min_value[ad] > self.max_value[ad] - p2[ad]:
                        alfa1[ad] = 1 + 2 * (self.max_value[ad] - p2[ad]) / (p2[ad] - p1[ad])
                    else:
                        alfa1[ad] = 1 + 2 * (p1[ad] - self.min_value[ad]) / (p2[ad] - p1[ad])
                alfa1 = 1 / alfa1
                alfa1 = 2 - (alfa1 ** (self.nc + 1))
                alfa2 = alfa1
                alfa1 = np.where(np.isnan(alfa1), 1, alfa1)
                alfa2 = np.where(np.isnan(alfa2), 1, alfa2)
                randList = np.random.random(self.chrom_length)
                orTF = (randList <= 1 / alfa1)
                aq = np.array([0.0] * self.chrom_length)
                for j in range(self.chrom_length):
                    if orTF[j] == True:
                        aq[j] = (randList[j] * alfa1[j]) ** (1.0 / (self.nc + 1))
                    else:
                        expp = 2.0 - (randList[j] * alfa1[j])
                        aq[j] = (1.0 / expp) ** (1.0 / (self.nc + 1))

                randList = np.random.random(self.chrom_length)
                orTF = (randList <= 1 / alfa2)
                bq = np.array([0.0] * self.chrom_length)
                for j in range(self.chrom_length):
                    if orTF[j] == True:
                        bq[j] = (randList[j] * alfa2[j]) ** (1.0 / (self.nc + 1))
                    else:
                        expp = 2.0 - (randList[j] * alfa2[j])
                        bq[j] = (1.0 / expp) ** (1.0 / (self.nc + 1))
                tp1 = 0.5 * ((1 + aq) * p1 + (1 - aq) * p2)
                tp2 = 0.5 * ((1 - bq) * p1 + (1 + bq) * p2)
                tp1 = np.max(np.vstack((tp1, np.array([self.min_value] * self.chrom_length))), 0)
                tp1 = np.min(np.vstack((tp1, np.array([self.max_value] * self.chrom_length))), 0)
                tp2 = np.max(np.vstack((tp2, np.array([self.min_value] * self.chrom_length))), 0)
                tp2 = np.min(np.vstack((tp2, np.array([self.max_value] * self.chrom_length))), 0)
                self.pop = np.row_stack((self.pop, tp1, tp2))

    def mutation(self, pm):
        for i in range(self.pop_size):
            self.pop = np.row_stack((self.pop, self.pop[i]))
            for j in range(self.chrom_length):
                if (random.random() < pm):
                    mpoint = self.pop[i][j]
                    if (mpoint - self.min_value[j]) < (self.max_value[j] - mpoint):
                        temp = 1 - (mpoint - self.min_value[j]) / (self.max_value[j] - self.min_value[j])
                    else:
                        temp = 1 - (self.max_value[j] - mpoint) / (self.max_value[j] - self.min_value[j])
                    u = random.random()
                    if u <= 0.5:
                        delta = ((2 * u + (1 - 2 * u) * (temp ** (self.nm + 1))) ** (1 / (self.nm + 1))) - 1
                    else:
                        delta = 1 - ((2 * (1 - u) + (2 * u - 1) * (temp ** (self.nm + 1))) ** (1 / (self.nm + 1)))
                    self.pop[i][j] = mpoint + delta * (self.max_value[j] - self.min_value[j])



import warnings
from typing import Union

import gym
from gym import spaces
import numpy as np

from neorl.rl.baselines.shared.vec_env import DummyVecEnv, VecCheckNan


def _enforce_array_obs(observation_space: spaces.Space) -> bool:
    """
    Whether to check that the returned observation is a numpy array
    it is not mandatory for `Dict` and `Tuple` spaces.
    """
    return not isinstance(observation_space, (spaces.Dict, spaces.Tuple))


def _check_image_input(observation_space: spaces.Box) -> None:
    """
    Check that the input will be compatible with Stable-Baselines
    when the observation is apparently an image.
    """
    if observation_space.dtype != np.uint8:
        warnings.warn("It seems that your observation is an image but the `dtype` "
                      "of your observation_space is not `np.uint8`. "
                      "If your observation is not an image, we recommend you to flatten the observation "
                      "to have only a 1D vector")

    if np.any(observation_space.low != 0) or np.any(observation_space.high != 255):
        warnings.warn("It seems that your observation space is an image but the "
                      "upper and lower bounds are not in [0, 255]. "
                      "Because the CNN policy normalize automatically the observation "
                      "you may encounter issue if the values are not in that range."
                      )

    if observation_space.shape[0] < 36 or observation_space.shape[1] < 36:
        warnings.warn("The minimal resolution for an image is 36x36 for the default CnnPolicy. "
                      "You might need to use a custom `cnn_extractor` "
                      "cf https://stable-baselines.readthedocs.io/en/master/guide/custom_policy.html")


def _check_unsupported_obs_spaces(env: gym.Env, observation_space: spaces.Space) -> None:
    """Emit warnings when the observation space used is not supported by Stable-Baselines."""

    if isinstance(observation_space, spaces.Dict) and not isinstance(env, gym.GoalEnv):
        warnings.warn("The observation space is a Dict but the environment is not a gym.GoalEnv "
                      "(cf https://github.com/openai/gym/blob/master/gym/core.py), "
                      "this is currently not supported by Stable Baselines "
                      "(cf https://github.com/hill-a/stable-baselines/issues/133), "
                      "you will need to use a custom policy. "
                      )

    if isinstance(observation_space, spaces.Tuple):
        warnings.warn("The observation space is a Tuple,"
                      "this is currently not supported by Stable Baselines "
                      "(cf https://github.com/hill-a/stable-baselines/issues/133), "
                      "you will need to flatten the observation and maybe use a custom policy. "
                      )


def _check_nan(env: gym.Env) -> None:
    """Check for Inf and NaN using the VecWrapper."""
    vec_env = VecCheckNan(DummyVecEnv([lambda: env]))
    for _ in range(10):
        action = [env.action_space.sample()]
        _, _, _, _ = vec_env.step(action)


def _check_obs(obs: Union[tuple, dict, np.ndarray, int],
               observation_space: spaces.Space,
               method_name: str) -> None:
    """
    Check that the observation returned by the environment
    correspond to the declared one.
    """
    if not isinstance(observation_space, spaces.Tuple):
        assert not isinstance(obs, tuple), ("The observation returned by the `{}()` "
                                            "method should be a single value, not a tuple".format(method_name))

    # The check for a GoalEnv is done by the base class
    if isinstance(observation_space, spaces.Discrete):
        assert isinstance(obs, int), "The observation returned by `{}()` method must be an int".format(method_name)
    elif _enforce_array_obs(observation_space):
        assert isinstance(obs, np.ndarray), ("The observation returned by `{}()` "
                                             "method must be a numpy array".format(method_name))

    assert observation_space.contains(obs), ("The observation returned by the `{}()` "
                                             "method does not match the given observation space".format(method_name))


def _check_returned_values(env: gym.Env, observation_space: spaces.Space, action_space: spaces.Space) -> None:
    """
    Check the returned values by the env when calling `.reset()` or `.step()` methods.
    """
    # because env inherits from gym.Env, we assume that `reset()` and `step()` methods exists
    obs = env.reset()

    _check_obs(obs, observation_space, 'reset')

    # Sample a random action
    action = action_space.sample()
    data = env.step(action)

    assert len(data) == 4, "The `step()` method must return four values: obs, reward, done, info"

    # Unpack
    obs, reward, done, info = data

    _check_obs(obs, observation_space, 'step')

    # We also allow int because the reward will be cast to float
    assert isinstance(reward, (float, int)), "The reward returned by `step()` must be a float"
    assert isinstance(done, bool), "The `done` signal must be a boolean"
    assert isinstance(info, dict), "The `info` returned by `step()` must be a python dictionary"

    if isinstance(env, gym.GoalEnv):
        # For a GoalEnv, the keys are checked at reset
        assert reward == env.compute_reward(obs['achieved_goal'], obs['desired_goal'], info)


def _check_spaces(env: gym.Env) -> None:
    """
    Check that the observation and action spaces are defined
    and inherit from gym.spaces.Space.
    """
    # Helper to link to the code, because gym has no proper documentation
    gym_spaces = " cf https://github.com/openai/gym/blob/master/gym/spaces/"

    assert hasattr(env, 'observation_space'), "You must specify an observation space (cf gym.spaces)" + gym_spaces
    assert hasattr(env, 'action_space'), "You must specify an action space (cf gym.spaces)" + gym_spaces

    assert isinstance(env.observation_space,
                      spaces.Space), "The observation space must inherit from gym.spaces" + gym_spaces
    assert isinstance(env.action_space, spaces.Space), "The action space must inherit from gym.spaces" + gym_spaces


def _check_render(env: gym.Env, warn: bool = True, headless: bool = False) -> None:
    """
    Check the declared render modes and the `render()`/`close()`
    method of the environment.

    :param env: (gym.Env) The environment to check
    :param warn: (bool) Whether to output additional warnings
    :param headless: (bool) Whether to disable render modes
        that require a graphical interface. False by default.
    """
    render_modes = env.metadata.get('render.modes')
    if render_modes is None:
        if warn:
            warnings.warn("No render modes was declared in the environment "
                          " (env.metadata['render.modes'] is None or not defined), "
                          "you may have trouble when calling `.render()`")

    else:
        # Don't check render mode that require a
        # graphical interface (useful for CI)
        if headless and 'human' in render_modes:
            render_modes.remove('human')
        # Check all declared render modes
        for render_mode in render_modes:
            env.render(mode=render_mode)
        env.close()


def check_env(env: gym.Env, warn: bool = True, skip_render_check: bool = True) -> None:
    """
    Check that an environment follows Gym API.
    This is particularly useful when using a custom environment.
    Please take a look at https://github.com/openai/gym/blob/master/gym/core.py
    for more information about the API.

    It also optionally check that the environment is compatible with Stable-Baselines.

    :param env: (gym.Env) The Gym environment that will be checked
    :param warn: (bool) Whether to output additional warnings
        mainly related to the interaction with Stable Baselines
    :param skip_render_check: (bool) Whether to skip the checks for the render method.
        True by default (useful for the CI)
    """
    assert isinstance(env, gym.Env), ("You environment must inherit from gym.Env class "
                                      " cf https://github.com/openai/gym/blob/master/gym/core.py")

    # ============= Check the spaces (observation and action) ================
    _check_spaces(env)

    # Define aliases for convenience
    observation_space = env.observation_space
    action_space = env.action_space

    # Warn the user if needed.
    # A warning means that the environment may run but not work properly with Stable Baselines algorithms
    if warn:
        _check_unsupported_obs_spaces(env, observation_space)

        # If image, check the low and high values, the type and the number of channels
        # and the shape (minimal value)
        if isinstance(observation_space, spaces.Box) and len(observation_space.shape) == 3:
            _check_image_input(observation_space)

        if isinstance(observation_space, spaces.Box) and len(observation_space.shape) not in [1, 3]:
            warnings.warn("Your observation has an unconventional shape (neither an image, nor a 1D vector). "
                          "We recommend you to flatten the observation "
                          "to have only a 1D vector")

        # Check for the action space, it may lead to hard-to-debug issues
        if (isinstance(action_space, spaces.Box) and
                (np.any(np.abs(action_space.low) != np.abs(action_space.high))
                 or np.any(np.abs(action_space.low) > 1) or np.any(np.abs(action_space.high) > 1))):
            warnings.warn("We recommend you to use a symmetric and normalized Box action space (range=[-1, 1]) "
                          "cf https://stable-baselines.readthedocs.io/en/master/guide/rl_tips.html")

    # ============ Check the returned values ===============
    _check_returned_values(env, observation_space, action_space)

    # ==== Check the render method and the declared render modes ====
    if not skip_render_check:
        _check_render(env, warn=warn)

    # The check only works with numpy arrays
    if _enforce_array_obs(observation_space):
        _check_nan(env)



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#"""
#Created on Mon Jun 29 15:36:46 2020
#
#@author: alyssawang
#"""

import inspect
import numpy as np
import pandas as pd
import joblib
import matplotlib.pyplot as plt

# Scikit-optimise
from skopt import gp_minimize
from skopt.space import Integer, Real, Categorical
from skopt.utils import use_named_args

class BAYESTUNE:
    """
    A module for Bayesian search for hyperparameter tuning
    
    :param param_grid: (dict) the type and range of each hyperparameter in a dictionary form (types are ``int/discrete`` or ``float/continuous`` or ``grid/categorical``). Example: {'x1': [[40, 50, 60, 100], 'grid'], 'x2': [[0.2, 0.8], 'float'], 'x3': [['blend', 'cx2point'], 'grid'], 'x4': [[20, 80], 'int']}
    :param fit: (function) the self-defined fitness function that includes the hyperparameters as input and algorithm score as output
    :param mode: (str) problem type, either "min" for minimization problem or "max" for maximization. Default: Bayesian tuner is set to minimize an objective
    :param ncases: (int) number of random hyperparameter cases to generate per core, ``ncases >= 11`` (see **Notes** for an important remark) 
    :param seed: (int) random seed for sampling reproducibility
    """
    def __init__(self, param_grid, fit, mode='min', ncases=50, seed=None):
        self.mode=mode
        assert self.mode in ['min', 'max'], '--error: The mode entered by user is invalid, use either `min` or `max`'
        self.param_grid=param_grid
        self.fit=fit
        self.ncases=ncases
        self.seed=seed
        if self.ncases < 11:
            print('--warning: ncases={} < 11 is given by the user, but ncases must be more than 11, reset ncases to 11'.format(self.ncases))
            self.ncases = 11
            
        self.full_grid()
        
    def get_func_args(self, f):
        #this function returns the argument names of the input function "f"
        return inspect.getfullargspec(f)[0]
    
#    def plot_results(self):
#        plot_convergence(self.search_result)

    def full_grid(self):
        #This function parses the param_grid variable from the user and sets up the 
        #parameter space for Bayesian search
        
        self.param_types=[self.param_grid[item][0] for item in self.param_grid]
        self.param_lst=[]
        for i, item in enumerate(self.param_grid):
            if self.param_types[i] in ['grid', 'categorical']:
                self.param_lst.append(self.param_grid[item][1])
            else:
                self.param_lst.append(self.param_grid[item][1:])
        
        self.param_names=[item for item in self.param_grid]
        self.dimensions=[]
        self.func_args=self.get_func_args(self.fit)
        
        for types, vals, names in zip(self.param_types, self.param_lst, self.param_names):
            if types in ['int', 'discrete']:
                lb=vals[0]
                ub=vals[1]
                self.dimensions.append(Integer(low=lb, high=ub, name=names))
            elif types in ['float', 'continuous']:
                lb=vals[0]
                ub=vals[1]
                self.dimensions.append(Real(low=lb, high=ub, name=names))
            elif types in ['grid', 'categorical']: 
                real_grid=vals
                self.dimensions.append(Categorical(categories=tuple(real_grid),  name=names))
            else:
                raise Exception('--error: the param types must be one of int/discrete or float/continuous or grid/categorical, this type is not avaiable: `{}`'.format(types))
                
    def worker(self,x):
        #This function setup a case worker to pass to the Parallel pool
        
        if self.mode=='min':
            @use_named_args(dimensions=self.dimensions)
            def fitness_wrapper(*args, **kwargs):
                return self.fit(*args, **kwargs)             
        else:
            @use_named_args(dimensions=self.dimensions)
            def fitness_wrapper(*args, **kwargs):
                return -self.fit(*args, **kwargs) 
        
        if self.seed is not None:
            core_seed=self.seed + x
        else:
            core_seed=None

        search_result = gp_minimize(func=fitness_wrapper,
                                    dimensions=self.dimensions,
                                    acq_func='EI', # Expected Improvement.
                                    n_calls=self.ncases,
                                    random_state=core_seed, verbose=self.verbose)
        
        return search_result.x_iters, list(search_result.func_vals)
    
    def plot_results(self, pngname='bayes_tune'):
        if self.mode=='max':
            plt.plot(pd.DataFrame.cummax(self.bayesres['score']), '-og')
            plt.ylabel('Max score so far')
        else:
            plt.plot(pd.DataFrame.cummin(self.bayesres['score']), '-og')
            plt.ylabel('Min score so far')
            
        plt.xlabel('Iteration')
        plt.grid()
        if pngname is not None:
            plt.savefig(str(pngname)+'.png', dpi=200, format='png')
        plt.close()
        
    def tune(self, ncores=1, csvname=None, verbose=True):
        """
        This function starts the tuning process with specified number of processors
    
        :param nthreads: (int) number of parallel threads (see the **Notes** section below for an important note about parallel execution)
        :param csvname: (str) the name of the csv file name to save the tuning results (useful for expensive cases as the csv file is updated directly after the case is done)
        :param verbose: (bool) whether to print updates to the screen or not
        """
        self.ncores=ncores
        self.csvlogger=csvname
        self.verbose=verbose

        if self.verbose:
            print('***************************************************************')
            print('****************Bayesian Search is Running*********************')
            print('***************************************************************')
            
            if self.ncores > 1:
                print('--- Running in parallel with {} threads and {} cases per threads'.format(self.ncores, self.ncases))
                print('--- Total number of executed cases is {}*{}={} cases'.format(self.ncores,self.ncases,self.ncores*self.ncases))
   
        if self.ncores > 1:
            
            with joblib.Parallel(n_jobs=self.ncores) as parallel:
                x_vals, func_vals=zip(*parallel(joblib.delayed(self.worker)(core+1) for core in range(self.ncores)))
            
            #flatten the x-lists for all cores
            x_vals_flatten=[]
            for lists in x_vals:
                for item in lists:
                    x_vals_flatten.append(item)
            
            #flatten the y results from all cores 
            func_vals_flatten = [item for sublist in func_vals for item in sublist]

            assert len(func_vals_flatten) == len(x_vals_flatten), '--error: the length of func_vals_flatten and x_vals_flatten in parallel Bayesian search must be equal'
            self.bayesres=pd.DataFrame(x_vals_flatten, columns = self.func_args)
            
            self.bayesres['score'] = np.array(func_vals_flatten) if self.mode=='min' else -np.array(func_vals_flatten)
         
        else:
            
            if self.mode=='min':
                @use_named_args(dimensions=self.dimensions)
                def fitness_wrapper(*args, **kwargs):
                    return self.fit(*args, **kwargs)             
            else:
                @use_named_args(dimensions=self.dimensions)
                def fitness_wrapper(*args, **kwargs):
                    return -self.fit(*args, **kwargs) 
            
            #Single core search
            self.search_result = gp_minimize(func=fitness_wrapper,
                                            dimensions=self.dimensions,
                                            acq_func='EI', # Expected Improvement.
                                            n_calls=self.ncases,
                                            random_state=self.seed, verbose=self.verbose)

            self.bayesres = pd.DataFrame(self.search_result.x_iters, columns = self.func_args)
            self.bayesres['score'] = self.search_result.func_vals if self.mode=='min' else -self.search_result.func_vals

        self.bayesres.index+=1
        
        if self.csvlogger:
            self.bayesres.index.name='id'
            self.bayesres.to_csv(self.csvlogger)
                
        return self.bayesres


from neorl.tune import BAYESTUNE
from neorl import ES

def test_bayes():
    
    #Define the fitness function (for original optimisation)
    def sphere(individual):
        y=sum(x**2 for x in individual)
        return y
    
    def tune_fit(cxpb, mu, alpha, cxmode):
    
        #--setup the parameter space
        nx=5
        BOUNDS={}
        for i in range(1,nx+1):
                BOUNDS['x'+str(i)]=['float', -100, 100]
    
        #--setup the ES algorithm
        es=ES(mode='min', bounds=BOUNDS, fit=sphere, lambda_=80, mu=mu, mutpb=0.1, alpha=alpha,
                 cxmode=cxmode, cxpb=cxpb, ncores=1, seed=1)
    
        #--Evolute the ES object and obtains y_best
        #--turn off verbose for less algorithm print-out when tuning
        x_best, y_best, es_hist=es.evolute(ngen=100, verbose=0)
    
        return y_best #returns the best score
    
    #Setup the parameter space
    #VERY IMPORTANT: The order of these parameters MUST be similar to their order in tune_fit
    #see tune_fit
    param_grid={
    #def tune_fit(cxpb, mu, alpha, cxmode):
    'cxpb': ['float', 0.1, 0.9],             #cxpb is first (low=0.1, high=0.8, type=float/continuous)
    'mu':   ['int', 30, 60],                 #mu is second (low=30, high=60, type=int/discrete)
    'alpha':['grid', [0.1, 0.2, 0.3, 0.4]],    #alpha is third (grid with fixed values, type=grid/categorical)
    'cxmode':['grid', ['blend', 'cx2point']]}  #cxmode is fourth (grid with fixed values, type=grid/categorical)
    
    #setup a bayesian tune object
    btune=BAYESTUNE(mode='min', param_grid=param_grid, fit=tune_fit, ncases=30)
    #tune the parameters with method .tune
    bayesres=btune.tune(ncores=1, verbose=True)
    print(bayesres)

test_bayes()


import tensorflow as tf


def gmatmul(tensor_a, tensor_b, transpose_a=False, transpose_b=False, reduce_dim=None):
    """
    Do a matrix multiplication with tensor 'a' and 'b', even when their shape do not match

    :param tensor_a: (TensorFlow Tensor)
    :param tensor_b: (TensorFlow Tensor)
    :param transpose_a: (bool) If 'a' needs transposing
    :param transpose_b: (bool) If 'b' needs transposing
    :param reduce_dim: (int) the multiplication over the dim
    :return: (TensorFlow Tensor) a * b
    """
    assert reduce_dim is not None

    # weird batch matmul
    if len(tensor_a.get_shape()) == 2 and len(tensor_b.get_shape()) > 2:
        # reshape reduce_dim to the left most dim in b
        b_shape = tensor_b.get_shape()
        if reduce_dim != 0:
            b_dims = list(range(len(b_shape)))
            b_dims.remove(reduce_dim)
            b_dims.insert(0, reduce_dim)
            tensor_b = tf.transpose(tensor_b, b_dims)
        b_t_shape = tensor_b.get_shape()
        tensor_b = tf.reshape(tensor_b, [int(b_shape[reduce_dim]), -1])
        result = tf.matmul(tensor_a, tensor_b, transpose_a=transpose_a,
                           transpose_b=transpose_b)
        result = tf.reshape(result, b_t_shape)
        if reduce_dim != 0:
            b_dims = list(range(len(b_shape)))
            b_dims.remove(0)
            b_dims.insert(reduce_dim, 0)
            result = tf.transpose(result, b_dims)
        return result

    elif len(tensor_a.get_shape()) > 2 and len(tensor_b.get_shape()) == 2:
        # reshape reduce_dim to the right most dim in a
        a_shape = tensor_a.get_shape()
        outter_dim = len(a_shape) - 1
        reduce_dim = len(a_shape) - reduce_dim - 1
        if reduce_dim != outter_dim:
            a_dims = list(range(len(a_shape)))
            a_dims.remove(reduce_dim)
            a_dims.insert(outter_dim, reduce_dim)
            tensor_a = tf.transpose(tensor_a, a_dims)
        a_t_shape = tensor_a.get_shape()
        tensor_a = tf.reshape(tensor_a, [-1, int(a_shape[reduce_dim])])
        result = tf.matmul(tensor_a, tensor_b, transpose_a=transpose_a,
                           transpose_b=transpose_b)
        result = tf.reshape(result, a_t_shape)
        if reduce_dim != outter_dim:
            a_dims = list(range(len(a_shape)))
            a_dims.remove(outter_dim)
            a_dims.insert(reduce_dim, outter_dim)
            result = tf.transpose(result, a_dims)
        return result

    elif len(tensor_a.get_shape()) == 2 and len(tensor_b.get_shape()) == 2:
        return tf.matmul(tensor_a, tensor_b, transpose_a=transpose_a, transpose_b=transpose_b)

    assert False, 'something went wrong'


def clipout_neg(vec, threshold=1e-6):
    """
    clip to 0 if input lower than threshold value

    :param vec: (TensorFlow Tensor)
    :param threshold: (float) the cutoff threshold
    :return: (TensorFlow Tensor) clipped input
    """
    mask = tf.cast(vec > threshold, tf.float32)
    return mask * vec


def detect_min_val(input_mat, var, threshold=1e-6, name='', debug=False):
    """
    If debug is not set, will run clipout_neg. Else, will clip and print out odd eigen values

    :param input_mat: (TensorFlow Tensor)
    :param var: (TensorFlow Tensor) variable
    :param threshold: (float) the cutoff threshold
    :param name: (str) the name of the variable
    :param debug: (bool) debug function
    :return: (TensorFlow Tensor) clipped tensor
    """
    eigen_min = tf.reduce_min(input_mat)
    eigen_max = tf.reduce_max(input_mat)
    eigen_ratio = eigen_max / eigen_min
    input_mat_clipped = clipout_neg(input_mat, threshold)

    if debug:
        input_mat_clipped = tf.cond(tf.logical_or(tf.greater(eigen_ratio, 0.), tf.less(eigen_ratio, -500)),
                                    lambda: input_mat_clipped, lambda: tf.Print(
                input_mat_clipped,
                [tf.convert_to_tensor('odd ratio ' + name + ' eigen values!!!'), tf.convert_to_tensor(var.name),
                 eigen_min, eigen_max, eigen_ratio]))

    return input_mat_clipped


def factor_reshape(eigen_vectors, eigen_values, grad, fac_idx=0, f_type='act'):
    """
    factor and reshape input eigen values

    :param eigen_vectors: ([TensorFlow Tensor]) eigen vectors
    :param eigen_values: ([TensorFlow Tensor]) eigen values
    :param grad: ([TensorFlow Tensor]) gradient
    :param fac_idx: (int) index that should be factored
    :param f_type: (str) function type to factor and reshape
    :return: ([TensorFlow Tensor], [TensorFlow Tensor]) factored and reshaped eigen vectors
            and eigen values
    """
    grad_shape = grad.get_shape()
    if f_type == 'act':
        assert eigen_values.get_shape()[0] == grad_shape[fac_idx]
        expanded_shape = [1, ] * len(grad_shape)
        expanded_shape[fac_idx] = -1
        eigen_values = tf.reshape(eigen_values, expanded_shape)
    if f_type == 'grad':
        assert eigen_values.get_shape()[0] == grad_shape[len(grad_shape) - fac_idx - 1]
        expanded_shape = [1, ] * len(grad_shape)
        expanded_shape[len(grad_shape) - fac_idx - 1] = -1
        eigen_values = tf.reshape(eigen_values, expanded_shape)

    return eigen_vectors, eigen_values



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Thu Dec  3 14:42:29 2020
#
#@author: Katelin Du
#"""


# Harris hawks optimization: Algorithm and applications
# Ali Asghar Heidari, Seyedali Mirjalili, Hossam Faris, 
# Ibrahim Aljarah, Majdi Mafarja, Huiling Chen
# Future Generation Computer Systems,
# DOI: https://doi.org/10.1016/j.future.2019.02.028

import random
import numpy as np
import math
import time
import joblib
import matplotlib.pyplot as plt
from neorl.evolu.discrete import mutate_discrete, encode_grid_to_discrete 
from neorl.evolu.discrete import decode_discrete_to_grid, encode_grid_indv_to_discrete
from neorl.utils.seeding import set_neorl_seed
from neorl.utils.tools import get_population, check_mixed_individual

class HHO(object):
    """
    Harris Hawks Optimizer
    
    :param mode: (str) problem type, either "min" for minimization problem or "max" for maximization
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}``
    :param fit: (function) the fitness function 
    :param nhawks: (int): number of the hawks in the group
    :param int_transform: (str): method of handling int/discrete variables, choose from: ``nearest_int``, ``sigmoid``, ``minmax``.
    :param ncores: (int) number of parallel processors (must be ``<= nhawks``)
    :param seed: (int) random seed for sampling
    """
    def __init__(self, mode, bounds, fit, nhawks, int_transform='nearest_int', ncores=1, seed=None):
        
        self.seed = seed
        set_neorl_seed(self.seed)

        #--mir
        assert mode == 'min' or mode == 'max', "Mode must be 'max' or 'min'."
        self.mode=mode
        if mode == 'min':
            self.fit=fit
        elif mode == 'max':
            def fitness_wrapper(*args, **kwargs):
                return -fit(*args, **kwargs) 
            self.fit=fitness_wrapper
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')
            
        self.int_transform=int_transform
        self.ncores = ncores
        self.nhawks = nhawks
        self.dim = len(bounds)
        self.bounds = bounds
        
        #infer variable types 
        self.var_type = np.array([bounds[item][0] for item in bounds])
        
        #mir-grid
        if "grid" in self.var_type:
            self.grid_flag=True
            self.orig_bounds=bounds  #keep original bounds for decoding
            #print('--debug: grid parameter type is found in the space')
            self.bounds, self.bounds_map=encode_grid_to_discrete(self.bounds) #encoding grid to int
            #define var_types again by converting grid to int
            self.var_type = np.array([self.bounds[item][0] for item in self.bounds])
        else:
            self.grid_flag=False
            self.bounds = bounds
            self.orig_bounds=bounds
        
        self.lb = np.array([self.bounds[item][1] for item in self.bounds])
        self.ub = np.array([self.bounds[item][2] for item in self.bounds])
        
    def evolute(self, ngen, x0=None, verbose=False, **kwargs):
        """
        This function evolutes the HHO algorithm for number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param x0: (list of lists) initial position of the hawks (must be of same size as ``nhawks``)
        :param verbose: (bool) print statistics to screen
        
        :return: (tuple) (best individual, best fitness, and dictionary containing major search results)
        """
        set_neorl_seed(self.seed)

        self.history = {'local_fitness':[], 'global_fitness':[]}
        self.rabbit_energy = float("inf")
        self.rabbit_location = np.zeros(self.dim)
        self.verbose = verbose

        ##################################
        # Set initial locations of hawks #
        ##################################
        self.hawk_positions = np.zeros((self.nhawks, self.dim))
        if x0:
            assert len(x0) == self.nhawks, '--error: the length of x0 ({}) MUST equal the number of hawks in the group ({})'.format(len(x0), self.nhawks)
            for i in range(self.nhawks):
                check_mixed_individual(x=x0[i], bounds=self.orig_bounds) #assert the type provided is consistent
                if self.grid_flag:
                    self.hawk_positions[i,:] = encode_grid_indv_to_discrete(x0[i], bounds=self.orig_bounds, bounds_map=self.bounds_map)
                else:
                    self.hawk_positions[i,:] = x0[i]
        else:
            # self.hawk_positions = np.asarray([x * (self.ub - self.lb) + self.lb for x in np.random.uniform(0, 1, (self.nhawks, self.dim))])
            for hawk_i in range(self.nhawks):
                self.hawk_positions[hawk_i, :] = self.init_sample()
        
        self.best_scores=[]
        for t in range(ngen):
            self.a= 1 - t * ((1) / ngen)  #mir: a decreases linearly between 1 to 0, for discrete mutation

            if 'E1' in kwargs:
                #take it from external vector
                assert len(kwargs["E1"]) == ngen, '--error: the length of `E1` in kwargs must equal to ngen'
                self.E1=kwargs["E1"][t]
            else:
                #E1 is annealed from 2 to 0
                self.E1 = 2 * (1 - (t / ngen))  # factor to show the decreasing energy of rabbit
            
            ###########################
            # Evaluate hawk fitnesses #
            ###########################
            fitness_lst = self.eval_hawks()
            
            #for logging
            self.prev_pop=self.hawk_positions.copy()
            self.prev_fits=np.array(fitness_lst)
            self.best_scores.append(np.min(fitness_lst))
            #######################################################################
            # Update rabbit energy and rabbit location based on best hawk fitness #
            #######################################################################
            for i, fitness in enumerate(fitness_lst):
                if fitness < self.rabbit_energy:
                    self.rabbit_energy = fitness
                    self.rabbit_location = self.hawk_positions[i, :].copy()
            
            #####################################################
            # Update best global and local fitnesses in history #
            #####################################################
            if self.mode=='max':
                self.best_global_fitness = -self.rabbit_energy
            else:
                self.best_global_fitness = self.rabbit_energy

            if self.verbose and t % self.nhawks: # change depending on how often message should be displayed
                print(f'HHO step {t*self.nhawks}/{ngen*self.nhawks}, nhawks={self.nhawks}, ncores={self.ncores}')
                print('Best global fitness:', np.round(self.best_global_fitness, 6))
                #mir-grid
                if self.grid_flag:
                    self.rabbit_decoded=decode_discrete_to_grid(self.rabbit_location,self.orig_bounds,self.bounds_map)
                    print('Best rabbit position:', self.rabbit_decoded)
                else:    
                    print('Best rabbit position:', np.round(self.rabbit_location, 6))
                print('E1:', self.E1)
                print()

            ################################
            # Update the location of hawks #
            ################################
            self.update_hawks(fitness_lst) # now self.hawk_positions is updated

            for hawk_i in range(self.nhawks):
                #mir: this bound check  line is needed to ensure that choices.remove option to work 
                self.hawk_positions[hawk_i, :] = self.ensure_bounds(self.hawk_positions[hawk_i, :], self.bounds)
                for dim in range(self.dim):
                    if self.var_type[dim] == 'int':
                        self.hawk_positions[hawk_i, dim] = mutate_discrete(x_ij=self.hawk_positions[hawk_i, dim], 
                                           x_min=self.hawk_positions[hawk_i, :].min(), 
                                           x_max=self.hawk_positions[hawk_i, :].max(),
                                           lb=self.lb[dim], 
                                           ub=self.ub[dim], 
                                           alpha=self.a,
                                           method=self.int_transform)

        #mir-grid
        if self.grid_flag:
            self.rabbit_correct=decode_discrete_to_grid(self.rabbit_location,self.orig_bounds,self.bounds_map)
        else:    
            self.rabbit_correct=self.rabbit_location
            
        if self.verbose:
            print('------------------------ HHO Summary --------------------------')
            print('Function:', self.fit.__name__)
            print('Best fitness (y) found:', self.best_global_fitness)
            print('Best individual (x) found:', self.rabbit_correct)
            print('-------------------------------------------------------------- \n \n')
        
        if self.mode=='max':
            self.prev_fits=-self.prev_fits
            self.best_scores=[-item for item in self.best_scores]
            self.history['global_fitness'] = np.maximum.accumulate(self.best_scores)
        else:
            self.history['global_fitness'] = np.minimum.accumulate(self.best_scores)
            
        self.history['local_fitness'] = self.best_scores
                
        #--mir return the last population for restart calculations
        if self.grid_flag:
            self.history['last_pop'] = get_population(self.prev_pop, fits=self.prev_fits, grid_flag=True, 
                                                     bounds=self.orig_bounds, bounds_map=self.bounds_map)
        else:
            self.history['last_pop'] = get_population(self.prev_pop, fits=self.prev_fits, grid_flag=False)
            
        
        return self.rabbit_correct, self.best_global_fitness, self.history

    def ensure_bounds(self, vec, bounds):
    
        vec_new = []
        # cycle through each variable in vector 
        for i, (key, val) in enumerate(bounds.items()):
    
            # variable exceedes the minimum boundary
            if vec[i] < bounds[key][1]:
                vec_new.append(bounds[key][1])
    
            # variable exceedes the maximum boundary
            if vec[i] > bounds[key][2]:
                vec_new.append(bounds[key][2])
    
            # the variable is fine
            if bounds[key][1] <= vec[i] <= bounds[key][2]:
                vec_new.append(vec[i])
            
        return vec_new

    def init_sample(self):
        #"""
        #Initialize a hawk location.

        #Return:
        #array - hawk position
        #"""
        hawk_pos = []
        for key in self.bounds:
            if self.bounds[key][0] == 'int':
                hawk_pos.append(random.randint(self.bounds[key][1], self.bounds[key][2]))
            elif self.bounds[key][0] == 'float':
                hawk_pos.append(random.uniform(self.bounds[key][1], self.bounds[key][2]))
            # elif self.bounds[key][0] == 'grid':
            #     hawk_pos.append(random.sample(self.bounds[key][1],1)[0])
            else:
                raise Exception ('unknown data type is given; either int, float, or grid are allowed for parameter bounds')
        return np.array(hawk_pos)

    def eval_hawks(self):
        #"""
        #Evaluate fitness of hawks with parallel processing.

        #Return:
        #list - hawk fitnesses
        #"""
        #print(self.hawk_positions)
        if self.ncores > 1:
            with joblib.Parallel(n_jobs=self.ncores) as parallel:
                fitness_lst = parallel(joblib.delayed(self.fit_worker)(self.hawk_positions[i, :]) for i in range(self.nhawks))
        else:
            fitness_lst = []
            for i in range(self.nhawks):
                fitness_lst.append(self.fit_worker(self.hawk_positions[i, :]))
        return fitness_lst

    def update_hawks(self, fitness_lst):
        #"""
        #Update hawk positions according to HHO rules.

        #Params:
        #cur_gen - current generation of HHO
        #ngen - total number of generations in HHO
        #fitness_lst - list of hawk fitnesses in current generation

        #Return:
        #None
        #"""            
        
        for i in range(self.nhawks):

            E0 = 2 * random.random() - 1
            rabbit_escape_energy = self.E1 * E0  # escaping energy of rabbit: eq. (3)

            ##############################
            # Exploration phase: eq. (1) #
            ##############################
            if abs(rabbit_escape_energy) >= 1:
                self.exploration(i)

            ###############################################################
            # Exploitation phase: attacking the rabbit using 4 strategies #
            ###############################################################
            elif abs(rabbit_escape_energy) < 1:
                r = random.random() # probablity of each event

                ##########################################
                # Phase 1: surprise pounce (seven kills) #
                ##########################################
                if r >= 0.5:
                    self.surprise_pounce(i, rabbit_escape_energy)

                #############################################################
                # Phase 2: performing team rapid dives (leapfrog movements) #
                #############################################################
                elif r < 0.5:
                    self.team_rapid_dives(i, fitness_lst[i], rabbit_escape_energy)

    def exploration(self, hawk_index):
        #"""
        #Exploration phase based on two perching strategies:
        #    1. new hawk location based on other hawks' locations
        #    2. new hawk location generated randomly

        #Params:
        #hawk_index - index of hawk being updated

        #Return:
        #None
        #"""
        q = random.random()
        rand_hawk = math.floor(self.nhawks * random.random())
        rand_hawk_pos = self.hawk_positions[rand_hawk, :] # location of random hawk

        if q < 0.5: # perch based on other family members
            self.hawk_positions[hawk_index, :] = rand_hawk_pos \
                                                - random.random() * abs(rand_hawk_pos - 2 * random.random() * self.hawk_positions[hawk_index, :])
        else: # perch on a random tall tree (random site inside group's home range)
            self.hawk_positions[hawk_index, :] = (self.rabbit_location - self.hawk_positions.mean(0)) \
                                                - random.random() * ((self.ub - self.lb) * random.random() + self.lb)

    def surprise_pounce(self, hawk_index, escape_energy):
        #"""
        #Exploitation phase 1 based on two strategies:
        #    1. Hard besiege without team rapid dives
        #    2. Soft besiege without team rapid dives

        #Params:
        #hawk_index - index of hawk being updated
        #escape_energy - escape energy of the rabbit

        #Return:
        #None
        #"""
        
        ########################
        # Hard besiege eq. (6) #
        ########################
        if abs(escape_energy) < 0.5:
            self.hawk_positions[hawk_index, :] = (self.rabbit_location) \
                                                - escape_energy * abs(self.rabbit_location - self.hawk_positions[hawk_index, :])
        ########################
        # Soft besiege eq. (4) #
        ########################
        elif abs(escape_energy) >= 0.5:
            J = 2 * (1 - random.random())  # random jump strength of the rabbit; 1 < J < 2
            #mir: the index was "i" here, which should be hawk_index
            self.hawk_positions[hawk_index, :] = (self.rabbit_location - self.hawk_positions[hawk_index, :]) \
                                        - escape_energy * abs(J * self.rabbit_location - self.hawk_positions[hawk_index, :])

    def team_rapid_dives(self, hawk_index, hawk_fitness, escape_energy):
        #"""
        #Exploitation phase 2 based on two strategies:
        #    1. Hard besiege with team rapid dives
        #    2. Soft besiege with team rapid dives

        #Params:
        #hawk_index - index of hawk being updated
        #hawk_fitness - fitness of current hawk
        #escape_energy - escape energy of the rabbit

        #Return:
        #None
        #"""
        #########################
        # Soft besiege eq. (10) #
        #########################
        if abs(escape_energy) >= 0.5:
            J = 2 * (1 - random.random())
            Y = self.rabbit_location \
                - escape_energy * abs(J * self.rabbit_location - self.hawk_positions[hawk_index, :])
            
            #mir: uses a built-in function 
            #(order is important for choices.remove(), first check bounds, then check discrete)
            Y=self.ensure_bounds(Y, self.bounds)
            Y=self.ensure_discrete(Y)
            
            if self.fit_worker(Y) < hawk_fitness:  # improved move?
                self.hawk_positions[hawk_index, :] = Y.copy()
            else:  # hawks perform self.levy-based short rapid dives around the rabbit
                Z = self.rabbit_location \
                    - escape_energy * abs(J * self.rabbit_location - self.hawk_positions[hawk_index, :]) \
                    + np.multiply(np.random.randn(self.dim), self.levy())
                    
                #mir---
                #(order is important for choices.remove(), first check bounds, then check discrete)
                Z=self.ensure_bounds(Z, self.bounds)
                Z=self.ensure_discrete(Z)
            
                if self.fit_worker(Z) < hawk_fitness:
                    self.hawk_positions[hawk_index, :] = Z.copy()
        #########################
        # Hard besiege eq. (11) #
        #########################
        elif abs(escape_energy) < 0.5:
            J = 2 * (1 - random.random())
            Y = self.rabbit_location \
                - escape_energy * abs(J * self.rabbit_location - self.hawk_positions.mean(0))
            
            #---mir
            #(order is important for choices.remove(), first check bounds, then check discrete)
            Y=self.ensure_bounds(Y, self.bounds)
            Y=self.ensure_discrete(Y)
        
            if self.fit_worker(Y) < hawk_fitness:  # improved move?
                self.hawk_positions[hawk_index, :] = Y.copy()
            else:  # Perform levy-based short rapid dives around the rabbit
                Z = self.rabbit_location \
                    - escape_energy * abs(J * self.rabbit_location - self.hawk_positions.mean(0)) \
                    + np.multiply(np.random.randn(self.dim), self.levy())
                    
                #--mir
                #(order is important for choices.remove(), first check bounds, then check discrete)
                Z=self.ensure_bounds(Z, self.bounds)
                Z=self.ensure_discrete(Z)
                
                if self.fit_worker(Z) < hawk_fitness:
                    self.hawk_positions[hawk_index, :] = Z.copy()
                    

    def fit_worker(self, hawk_pos):
        #"""
        #Evaluates fitness of a hawk.

        #Params:
        #hawk_pos - hawk position vector

        #Return:
        #float - hawk fitness
        #"""
        
        #mir---
        hawk_pos=self.ensure_bounds(hawk_pos, self.bounds)
        
        #mir-grid
        if self.grid_flag:
            #decode the individual back to the int/float/grid mixed space
            hawk_pos=decode_discrete_to_grid(hawk_pos,self.orig_bounds,self.bounds_map) 
                    
        fitness = self.fit(hawk_pos)
        return fitness
    
    def ensure_discrete(self, vec):
        #"""
        #to mutate a vector if discrete variables exist 
        #handy function to be used four times within HHO phases

        #Params:
        #vec - hawk position in vector/list form

        #Return:
        #vec - updated hawk position vector with discrete values
        #"""
        
        for dim in range(self.dim):
            if self.var_type[dim] == 'int':
                vec[dim] = mutate_discrete(x_ij=vec[dim], 
                                               x_min=min(vec),
                                               x_max=max(vec),
                                               lb=self.lb[dim], 
                                               ub=self.ub[dim],
                                               alpha=self.a,
                                               method=self.int_transform,
                                               )
        
        return vec

    def levy(self):
        #"""
        #Evaluates the levy flight (LF) function eq. (9).

        #Return:
        #float - result of levy flight function
        #"""
        beta = 1.5
        sigma = (math.gamma(1 + beta) * math.sin(math.pi * beta / 2) / (math.gamma((1 + beta) / 2) * beta * 2 ** ((beta - 1) / 2))) ** (1 / beta)
        u = 0.01 * np.random.randn(self.dim) * sigma
        v = np.random.randn(self.dim)
        zz = np.power(np.absolute(v), (1 / beta))
        step = np.divide(u, zz)
        return step



from neorl.benchmarks.classic import sphere, ackley, cigar

def test_sphere():
    y=sphere([0,0,0,0,0])
    assert y==0

def test_cigar():
    y=cigar([0]*10)
    assert y==0
    
def test_ackley():
    y=ackley([0]*100)
    assert y==0


import numpy as np

from neorl.rl.baselines.shared.vec_env import VecEnv


def evaluate_policy(model, env, n_eval_episodes=10, deterministic=True,
                    render=False, callback=None, reward_threshold=None,
                    return_episode_rewards=False):
    """
    Runs policy for `n_eval_episodes` episodes and returns average reward.
    This is made to work only with one env.

    :param model: (BaseRLModel) The RL agent you want to evaluate.
    :param env: (gym.Env or VecEnv) The gym environment. In the case of a `VecEnv`
        this must contain only one environment.
    :param n_eval_episodes: (int) Number of episode to evaluate the agent
    :param deterministic: (bool) Whether to use deterministic or stochastic actions
    :param render: (bool) Whether to render the environment or not
    :param callback: (callable) callback function to do additional checks,
        called after each step.
    :param reward_threshold: (float) Minimum expected reward per episode,
        this will raise an error if the performance is not met
    :param return_episode_rewards: (bool) If True, a list of reward per episode
        will be returned instead of the mean.
    :return: (float, float) Mean reward per episode, std of reward per episode
        returns ([float], [int]) when `return_episode_rewards` is True
    """
    if isinstance(env, VecEnv):
        assert env.num_envs == 1, "You must pass only one environment when using this function"

    episode_rewards, episode_lengths = [], []
    for _ in range(n_eval_episodes):
        obs = env.reset()
        done, state = False, None
        episode_reward = 0.0
        episode_length = 0
        while not done:
            action, state = model.predict(obs, state=state, deterministic=deterministic)
            obs, reward, done, _info = env.step(action)
            episode_reward += reward
            if callback is not None:
                callback(locals(), globals())
            episode_length += 1
            if render:
                env.render()
        episode_rewards.append(episode_reward)
        episode_lengths.append(episode_length)

    mean_reward = np.mean(episode_rewards)
    std_reward = np.std(episode_rewards)

    if reward_threshold is not None:
        assert mean_reward > reward_threshold, 'Mean reward below threshold: '\
                                         '{:.2f} < {:.2f}'.format(mean_reward, reward_threshold)
    if return_episode_rewards:
        return episode_rewards, episode_lengths
    return mean_reward, std_reward



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# Copyright (c) 2021, NEORL authors.
# Licensed under the MIT license
import warnings, os
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", message=r"Passing", category=FutureWarning)

import tensorflow as tf
from tensorflow.python.util import deprecation
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
if type(tf.contrib) != type(tf): tf.contrib._warning = None
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
deprecation._PRINT_DEPRECATION_WARNINGS = False

logo="""

\t    NEORL: NeuroEvolution Optimisation with Reinforcement Learning
\t\t\t ███╗   ██╗███████╗ ██████╗ ██████╗ ██╗     
\t\t\t ████╗  ██║██╔════╝██╔═══██╗██╔══██╗██║     
\t\t\t ██╔██╗ ██║█████╗  ██║   ██║██████╔╝██║     
\t\t\t ██║╚██╗██║██╔══╝  ██║   ██║██╔══██╗██║     
\t\t\t ██║ ╚████║███████╗╚██████╔╝██║  ██║███████╗
\t\t\t ╚═╝  ╚═══╝╚══════╝ ╚═════╝ ╚═╝  ╚═╝╚══════╝                                                
Copyright © 2021 Exelon Corporation (https://www.exeloncorp.com/) in collaboration with 
             MIT Nuclear Science and Engineering (https://web.mit.edu/nse/)
                             All Rights Reserved

                       \n"""
                       

try:                    
    print(logo)
except:
    print(logo.encode('utf-8'))
    #print(logo.encode('ascii', 'ignore').decode('ascii'))
import os

from neorl.rl.baselines.a2c import A2C
from neorl.rl.baselines.acer import ACER
from neorl.rl.baselines.deepq import DQN
from neorl.rl.baselines.ppo2 import PPO2
from neorl.rl.baselines.acktr import ACKTR
from neorl.evolu.pso import PSO
from neorl.evolu.sa import SA
from neorl.evolu.bat import BAT
from neorl.evolu.de import DE
from neorl.evolu.xnes import XNES
from neorl.evolu.es import ES
from neorl.evolu.gwo import GWO
from neorl.evolu.ssa import SSA
from neorl.evolu.woa import WOA
from neorl.evolu.jaya import JAYA
from neorl.evolu.mfo import MFO
from neorl.evolu.hho import HHO
from neorl.hybrid.pesa import PESA
from neorl.hybrid.pesa2 import PESA2
from neorl.rl.baselines.shared.policies import MlpPolicy
from neorl.rl.baselines.deepq.policies import MlpPolicy as DQNPolicy
from neorl.utils.neorlcalls import RLLogger
from neorl.rl.make_env import CreateEnvironment
from neorl.hybrid.rneat import RNEAT
from neorl.hybrid.fneat import FNEAT
from neorl.hybrid.ppoes import PPOES
from neorl.hybrid.ackde import ACKDE
from neorl.evolu.aco import ACO
from neorl.evolu.cs import CS
from neorl.hybrid.nga import NGA
from neorl.hybrid.nhho import NHHO
from neorl.evolu.ts import TS
from neorl.hybrid.aeo import AEO
from neorl.evolu.hclpso import HCLPSO
from neorl.hybrid.edev import EDEV
from neorl.hybrid.epso import EPSO


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 26 09:33:45 2020

@author: majdi
"""

def evaluate_policy(model, env, log_dir, n_eval_episodes=10, render=False):
    """
    test policy for `n_eval_episodes` episodes and returns reward.
    This is made to work only with one env and single core.
    :param model: The RL agent you want to evaluate.
    :param env: The gym environment.
    :param n_eval_episodes: (int) Number of episode to evaluate the agent
    :param render: (bool) Whether to render the environment or not
    """
    
    """
    if video_record:
        ffmpeg_code=os.system('which ffmpeg')
        if ffmpeg_code == 0: 
            print('--debug: ffmpeg is detected on the machine')
            ffmpeg=subprocess.check_output(['which', 'ffmpeg'])
            ffmpeg.decode('utf-8').strip()
        else:
            raise('The user activated video recording but ffmpeg is not installed on the machine')
    """
    
    episode_rewards, episode_lengths = [], []
    for _ in range(n_eval_episodes):
        obs = env.reset()
        done, state = False, None
        episode_reward = 0.0
        episode_length = 0
        while not done:
            action, state = model.predict(obs, state=state)
            obs, reward, done, _info = env.step(action)
            episode_reward += reward
            episode_length += 1
            #if render:
            #    env.render()
            
        episode_rewards.append(episode_reward)
        episode_lengths.append(episode_length)


#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#"""
#Created on Wed Feb 26 09:33:45 2020
#
#@author: majdi
#"""

import numpy as np
import random

def set_neorl_seed(seed=None):
    """
    This function seeds the NEORL algorthim for reproducibility
    :param seed: (int) value of random seed
    """
    
    if not (seed is None):
        random.seed(seed) 
        np.random.seed(seed)


import numpy as np
import scipy.signal


def safe_mean(arr):
    """
    Compute the mean of an array if there is at least one element.
    For empty array, return nan. It is used for logging only.

    :param arr: (np.ndarray)
    :return: (float)
    """
    return np.nan if len(arr) == 0 else np.mean(arr)


def discount(vector, gamma):
    """
    computes discounted sums along 0th dimension of vector x.
        y[t] = x[t] + gamma*x[t+1] + gamma^2*x[t+2] + ... + gamma^k x[t+k],
                where k = len(x) - t - 1

    :param vector: (np.ndarray) the input vector
    :param gamma: (float) the discount value
    :return: (np.ndarray) the output vector
    """
    assert vector.ndim >= 1
    return scipy.signal.lfilter([1], [1, -gamma], vector[::-1], axis=0)[::-1]


def explained_variance(y_pred, y_true):
    """
    Computes fraction of variance that ypred explains about y.
    Returns 1 - Var[y-ypred] / Var[y]

    interpretation:
        ev=0  =>  might as well have predicted zero
        ev=1  =>  perfect prediction
        ev<0  =>  worse than just predicting zero

    :param y_pred: (np.ndarray) the prediction
    :param y_true: (np.ndarray) the expected value
    :return: (float) explained variance of ypred and y
    """
    assert y_true.ndim == 1 and y_pred.ndim == 1
    var_y = np.var(y_true)
    return np.nan if var_y == 0 else 1 - np.var(y_true - y_pred) / var_y


def explained_variance_2d(y_pred, y_true):
    """
    Computes fraction of variance that ypred explains about y, for 2D arrays.
    Returns 1 - Var[y-ypred] / Var[y]

    interpretation:
        ev=0  =>  might as well have predicted zero
        ev=1  =>  perfect prediction
        ev<0  =>  worse than just predicting zero

    :param y_pred: (np.ndarray) the prediction
    :param y_true: (np.ndarray) the expected value
    :return: (float) explained variance of ypred and y
    """
    assert y_true.ndim == 2 and y_pred.ndim == 2
    var_y = np.var(y_true, axis=0)
    explained_var = 1 - np.var(y_true - y_pred) / var_y
    explained_var[var_y < 1e-10] = 0
    return explained_var


def flatten_arrays(arrs):
    """
    flattens a list of arrays down to 1D

    :param arrs: ([np.ndarray]) arrays
    :return: (np.ndarray) 1D flattened array
    """
    return np.concatenate([arr.flat for arr in arrs])


def unflatten_vector(vec, shapes):
    """
    reshape a flattened array

    :param vec: (np.ndarray) 1D arrays
    :param shapes: (tuple)
    :return: ([np.ndarray]) reshaped array
    """
    i = 0
    arrs = []
    for shape in shapes:
        size = np.prod(shape)
        arr = vec[i:i + size].reshape(shape)
        arrs.append(arr)
        i += size
    return arrs


def discount_with_boundaries(rewards, episode_starts, gamma):
    """
    computes discounted sums along 0th dimension of x (reward), while taking into account the start of each episode.
        y[t] = x[t] + gamma*x[t+1] + gamma^2*x[t+2] + ... + gamma^k x[t+k],
                where k = len(x) - t - 1

    :param rewards: (np.ndarray) the input vector (rewards)
    :param episode_starts: (np.ndarray) 2d array of bools, indicating when a new episode has started
    :param gamma: (float) the discount factor
    :return: (np.ndarray) the output vector (discounted rewards)
    """
    discounted_rewards = np.zeros_like(rewards)
    n_samples = rewards.shape[0]
    discounted_rewards[n_samples - 1] = rewards[n_samples - 1]
    for step in range(n_samples - 2, -1, -1):
        discounted_rewards[step] = rewards[step] + gamma * discounted_rewards[step + 1] * (1 - episode_starts[step + 1])
    return discounted_rewards


def scale_action(action_space, action):
    """
    Rescale the action from [low, high] to [-1, 1]
    (no need for symmetric action space)

    :param action_space: (gym.spaces.box.Box)
    :param action: (np.ndarray)
    :return: (np.ndarray)
    """
    low, high = action_space.low, action_space.high
    return 2.0 * ((action - low) / (high - low)) - 1.0


def unscale_action(action_space, scaled_action):
    """
    Rescale the action from [-1, 1] to [low, high]
    (no need for symmetric action space)

    :param action_space: (gym.spaces.box.Box)
    :param action: (np.ndarray)
    :return: (np.ndarray)
    """
    low, high = action_space.low, action_space.high
    return low + (0.5 * (scaled_action + 1.0) * (high - low))



#    This file is part of NEORL.

#    Copyright (c) 2021 Exelon Corporation and MIT Nuclear Science and Engineering
#    NEORL is free software: you can redistribute it and/or modify
#    it under the terms of the MIT LICENSE

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#    SOFTWARE.

# -*- coding: utf-8 -*-
#"""
#Created on Thu Dec  3 14:42:29 2020
#
#@author: Majdi
#"""

import random
import numpy as np
import joblib, sys
from numpy import arange, multiply, zeros, copy, ones, exp, dot, ceil
from numpy.random import uniform
from neorl.evolu.discrete import mutate_discrete, encode_grid_to_discrete 
from neorl.evolu.discrete import decode_discrete_to_grid, encode_grid_indv_to_discrete
from neorl.utils.seeding import set_neorl_seed
from neorl.utils.tools import get_population, check_mixed_individual
import pandas as pd

# Description 
# CLPSO/Sa-PSO(PSO,FDR,HPSO,LIPS,CLPSO-gbest)
# CLPSO: c1=3~1.5, w1=0.9~0.4
# Sa-PSO: PSO/FDR/HPSO/LIPS/CLPSO-gbest, w=0.9~0.2
# PSO: c2_1=2.5~0.5, c2_2=0.5~2.5
# HPSO: c3_1=2.5~0.5, c3_2=0.5~2.5
# LIPS: nsize=3
# CLPSO with gbest:c4_1=2.5~0.5, c4_2=0.5~2.5 & Pc=0.5

class EPSO(object):
    """
    Ensemble Particle Swarm Optimization (EPSO)
    
    :param mode: (str) problem type, either ``min`` for minimization problem or ``max`` for maximization
    :param bounds: (dict) input parameter type and lower/upper bounds in dictionary form. Example: ``bounds={'x1': ['int', 1, 4], 'x2': ['float', 0.1, 0.8], 'x3': ['float', 2.2, 6.2]}``
    :param fit: (function) the fitness function 
    :param g1: (int): number of particles in the exploration group
    :param g2: (int): number of particles in the exploitation group (total swarm size is ``g1 + g2``)
    :param int_transform: (str): method of handling int/discrete variables, choose from: ``nearest_int``, ``sigmoid``, ``minmax``.
    :param ncores: (int) number of parallel processors (must be ``<= g1+g2``)
    :param seed: (int) random seed for sampling
    """
    def __init__(self, mode, bounds, fit, g1=15, g2=25, int_transform='nearest_int', ncores=1, seed=None):
        
        set_neorl_seed(seed)
        
        assert ncores <= (g1+g2), '--error: ncores ({}) must be less than or equal to total particles g1 + g2 ({})'.format(ncores, g1+g2)
        
        #--mir
        self.mode=mode
        if mode == 'min':
            self.fit=fit
        elif mode == 'max':
            def fitness_wrapper(*args, **kwargs):
                return -fit(*args, **kwargs) 
            self.fit=fitness_wrapper
        else:
            raise ValueError('--error: The mode entered by user is invalid, use either `min` or `max`')
          
        self.int_transform=int_transform
        if int_transform not in ["nearest_int", "sigmoid", "minmax"]:
            raise ValueError('--error: int_transform entered by user is invalid, must be `nearest_int`, `sigmoid`, or `minmax`')
            
        self.bounds=bounds
        self.ncores = ncores
        self.num_g1=g1
        self.num_g2=g2
        self.num_g=self.num_g1 + self.num_g2
        
        #infer variable types 
        self.var_type = np.array([bounds[item][0] for item in bounds])
        
        #mir-grid
        if "grid" in self.var_type:
            self.grid_flag=True
            self.orig_bounds=bounds  #keep original bounds for decoding
            #print('--debug: grid parameter type is found in the space')
            self.bounds, self.bounds_map=encode_grid_to_discrete(self.bounds) #encoding grid to int
            #define var_types again by converting grid to int
            self.var_type = np.array([self.bounds[item][0] for item in self.bounds])
        else:
            self.grid_flag=False
            self.bounds = bounds
            self.orig_bounds=bounds
        
        self.dim = len(bounds)
        self.lb=np.array([self.bounds[item][1] for item in self.bounds])
        self.ub=np.array([self.bounds[item][2] for item in self.bounds])

    def init_sample(self, bounds):
    
        indv=[]
        for key in bounds:
            if bounds[key][0] == 'int':
                indv.append(random.randint(bounds[key][1], bounds[key][2]))
            elif bounds[key][0] == 'float':
                indv.append(random.uniform(bounds[key][1], bounds[key][2]))
            #elif bounds[key][0] == 'grid':
            #    indv.append(random.sample(bounds[key][1],1)[0])
            else:
                raise Exception ('unknown data type is given, either int, float, or grid are allowed for parameter bounds')   
        return np.array(indv)

    def eval_particles(self, pop):
    
        #---------------------
        # Fitness calcs
        #---------------------
        core_lst=[]
        for case in range (0, pop.shape[0]):
            core_lst.append(pop[case, :])
    
        if self.ncores > 1:

            with joblib.Parallel(n_jobs=self.ncores) as parallel:
                fitness_lst=parallel(joblib.delayed(self.fit_worker)(item) for item in core_lst)
                
        else:
            fitness_lst=[]
            for item in core_lst:
                fitness_lst.append(self.fit_worker(item))
        
        return fitness_lst

    def select(self, pos, fit):
        
        best_fit=np.min(fit)
        min_idx=np.argmin(fit)
        best_pos=pos[min_idx,:].copy()
        
        return best_pos, best_fit
        
    def ensure_bounds(self, vec):
    
        vec_new = []
        # cycle through each variable in vector 
        for i, (key, val) in enumerate(self.bounds.items()):
    
            # variable exceedes the minimum boundary
            if vec[i] < self.bounds[key][1]:
                vec_new.append(self.bounds[key][1])
    
            # variable exceedes the maximum boundary
            if vec[i] > self.bounds[key][2]:
                vec_new.append(self.bounds[key][2])
    
            # the variable is fine
            if self.bounds[key][1] <= vec[i] <= self.bounds[key][2]:
                vec_new.append(vec[i])
            
        return vec_new
    
    def fit_worker(self, x):
        #This worker is for parallel calculations
        
        # Clip the salp with position outside the lower/upper bounds and return same position
        x=self.ensure_bounds(x)
        
        if self.grid_flag:
            #decode the individual back to the int/float/grid mixed space
            x=decode_discrete_to_grid(x,self.orig_bounds,self.bounds_map)
        
        # Calculate objective function for each search agent
        fitness = self.fit(x)
        
        self.fitcount+=1
        
        return fitness
    
    def ensure_discrete(self, vec):
        #"""
        #to mutate a vector if discrete variables exist 
        #handy function to be used within alg phases

        #Params:
        #vec - salp position in vector/list form

        #Return:
        #vec - updated salp position vector with discrete values
        #"""
        
        for dim in range(self.dim):
            if self.var_type[dim] == 'int':
                vec[dim] = mutate_discrete(x_ij=vec[dim], 
                                               x_min=min(vec),
                                               x_max=max(vec),
                                               lb=self.lb[dim], 
                                               ub=self.ub[dim],
                                               alpha=self.a,
                                               method=self.int_transform,
                                               )
        
        return vec

    def UpdateParticles(self, a, b, friend_num, check_slope=False):
        
        #first particle index,
        #last particle index
        #number of particles to sample a friend
        
        for i in range(a,b):

            if check_slope:
                slope_cond=self.obj_func_slope[i] > 5
            else:
                slope_cond=True
                    
            if slope_cond:
                self.fri_best[i,:]=dot(i,ones((1,self.dim)))
                friend1=ceil(dot(friend_num,np.random.uniform(size=(self.dim)))) - 1
                friend2=ceil(dot(friend_num,np.random.uniform(size=(self.dim)))) - 1
                friend=multiply((self.pbest_val[friend1.astype(int)] < self.pbest_val[friend2.astype(int)]),friend1) \
                    + multiply((self.pbest_val[friend1.astype(int)] >= self.pbest_val[friend2.astype(int)]),friend2)
                toss=ceil(np.random.uniform(size=(self.dim)) - self.Pc[:,i].T)
                if np.all(toss == ones((self.dim))):
                    temp_index=np.random.choice(range(self.dim), self.dim, replace=False)
                    toss[temp_index[0]]=0
                self.fri_best[i,:]=multiply((1 - toss),friend) + multiply(toss,self.fri_best[i,:])
                for d in range(self.dim):
                    self.fri_best_pos[i,d]=self.pbest_pos[int(self.fri_best[i,d]),d]
                    
                if check_slope:
                    self.obj_func_slope[i]=0
            

    def evolute(self, ngen, LP=3, x0=None, verbose=False):
        """
        This function evolutes the EPSO algorithm for a number of generations.
        
        :param ngen: (int) number of generations to evolute
        :param LP: (int) number of generations before updating the success and failure memories for the ensemble variants (i.e. learning period)
        :param x0: (list of lists) initial position of the particles (must be of same size as ``g1 + g2``)
        :param verbose: (bool) print statistics to screen
        
        :return: (tuple) (best individual, best fitness, and dictionary containing major search results)
        """
        self.fitcount=0
        max_FES=int(ngen*self.num_g + self.num_g)   #maximum number of function evaluations for EPSO
        
        num_strategy=5  #number of variants
        success_hist=np.zeros((ngen,num_strategy))
        self.history = {'local_fitness':[], 'global_fitness':[]}
        self.best_fitness=float("inf") 
        self.verbose=verbose
        self.Positions = np.zeros((self.num_g, self.dim))
        self.a=1
        if x0:
            assert len(x0) == self.num_g, '--error: the length of x0 ({}) MUST equal the number of particles in the group `g1+g2 `({})'.format(len(x0), self.num_g)
            for i in range(self.num_g):
                check_mixed_individual(x=x0[i], bounds=self.orig_bounds) #assert the type provided is consistent
                if self.grid_flag:
                    self.Positions[i,:] = encode_grid_indv_to_discrete(x0[i], bounds=self.orig_bounds, bounds_map=self.bounds_map)
                else:
                    self.Positions[i,:] = x0[i]
        else:

            for i in range(self.num_g):
                self.Positions[i,:]=self.init_sample(self.bounds)
        
        # Initialization
        self.range_min=np.tile(self.lb,(self.num_g,1))
        self.range_max=np.tile(self.ub,(self.num_g,1))
        interval=self.range_max - self.range_min
        v_max=interval * 0.5
        v_min=-v_max
        self.vel=v_min + multiply((v_max - v_min),np.random.uniform(size=(self.num_g,self.dim)))
        
        #ensure discrete mutation
        for iii in range(self.Positions.shape[0]):
            self.Positions[iii, :] = self.ensure_bounds(self.Positions[iii, :])
            self.Positions[iii, :] = self.ensure_discrete(self.Positions[iii, :])
                
        fitness0=self.eval_particles(self.Positions)
        self.gbest_pos, self.gbest_val = self.select(self.Positions, fitness0)
        
        self.pbest_pos=self.Positions.copy()
        self.pbest_val=np.array(fitness0)
        
        # Method: CLPSO
        c1=3 - arange(ngen) * 1.5 / ngen
        w1=0.9 - arange(ngen) * (0.5 / ngen)
        
        # Sa-PSO (PSO/FDR/HPSO/LIPS/CLPSO-gbest)
        w2=0.9 - arange(ngen) * (0.7 / ngen)
        
        # Method 1: PSO
        c2_1=2.5 - arange(ngen) * 2 / ngen
        c2_2=0.5 + arange(ngen) * 2 / ngen
        
        # Method 2: FDR_PSO
        fii=np.array([1,1,2])
        
        # Method 3: HPSO_TVAC
        c3_1=2.5 - arange(ngen) * 2 / ngen
        c3_2=0.5 + arange(ngen) * 2 / ngen
        
        re_init_vel=np.zeros((ngen,self.dim))
        for di in range(self.dim):
            re_init_vel[:,di]=self.ub[di] - arange(ngen) * (0.9*self.ub[di]) / ngen
        
        # Method 4: LIPS
        nsize=3
        
        # Method 5: CLPSO
        c4_1=2.5 - arange(ngen) * 2 / ngen
        c4_2=0.5 + arange(ngen) * 2 / ngen
        
        j=np.linspace(0,1,self.num_g)
        j=j * 10
        self.Pc=np.dot(ones((self.dim,1)),(0.0 + (multiply((0.5),(exp(j)- exp(j[0]))) / (exp(j[self.num_g-1]) - exp(j[0]))))[np.newaxis])

        
        self.fri_best_pos=zeros((self.num_g,self.dim))
        self.fri_best=dot((arange(self.num_g)[np.newaxis]).T,ones((1,self.dim)))
        self.obj_func_slope=zeros((self.num_g))
        
        #Updating particles for group 1 (exploration)
        self.UpdateParticles(a=0, b=self.num_g1, friend_num=self.num_g1, check_slope=False)

        #Updating particles for group 2 (exploitation)
        self.UpdateParticles(a=self.num_g1, b=self.num_g, friend_num=self.num_g, check_slope=False) 
        
        
        delta=self.Positions.copy() * 0  #delta for PSO calcs
        update_flag=False       #a flag that will be switched once k mod LP is zero
        
        
        for k in range(ngen):
            
            self.a= 1 - k * ((1) / ngen)  #mir: a decreases linearly between 1 to 0, for discrete mutation
            
            gbest_pos_temp=np.tile(self.gbest_pos,(self.num_g,1))
            result=[]
            strat=[]
            
            #----------------------------------------------
            # Phase 1: Exploration
            #----------------------------------------------
            if self.fitcount <= 0.9 * max_FES:
                #print('--- Gen %i: EPSO Exploration Phase'%k)
                
                # Group 1: CLPSO
                self.UpdateParticles(a=0, b=self.num_g1, friend_num=self.num_g1, check_slope=True)
                
                delta[:self.num_g1,:]=multiply(c1[k]*uniform(size=(self.num_g1,self.dim)),(self.fri_best_pos[:self.num_g1,:] - self.Positions[:self.num_g1,:]))
                vel_g1=w1[k]*self.vel[:self.num_g1,:] + delta[:self.num_g1,:]
                self.vel[:self.num_g1,:]=(multiply((vel_g1 < v_min[:self.num_g1,:]),v_min[:self.num_g1,:]))  \
                            + (multiply((vel_g1 > v_max[:self.num_g1,:]),v_max[:self.num_g1,:])) \
                            + (multiply((np.logical_and((vel_g1 < v_max[:self.num_g1,:]),(vel_g1 > v_min[:self.num_g1,:]))),vel_g1))        
                
                self.Positions[:self.num_g1,:]=self.Positions[:self.num_g1,:] + self.vel[:self.num_g1,:]
                
                #for discrete mutation
                for iii in range(self.num_g1):
                    self.Positions[iii, :] = self.ensure_bounds(self.Positions[iii, :])
                    self.Positions[iii, :] = self.ensure_discrete(self.Positions[iii, :])
            
                result=self.eval_particles(pop=self.Positions[:self.num_g1,:])
                
                for i in range(self.num_g1):
                    
                    if result[i] < self.pbest_val[i]:
                        self.pbest_pos[i,:]=self.Positions[i,:].copy()
                        self.pbest_val[i]=result[i]
                        self.obj_func_slope[i]=0
                    else:
                        self.obj_func_slope[i]=self.obj_func_slope[i] + 1
                        
                    if self.pbest_val[i] < self.gbest_val:
                        self.gbest_pos=self.pbest_pos[i,:].copy()
                        self.gbest_val=self.pbest_val[i]
                        
                            
                for i in range(self.num_g1, self.num_g):
                    
                    ###########################################################
                    #DONT FORGET TO CHANGE IN EXPLOITATION PHASE.
                    #every line below this is a copy of what is in exploitation
                    ###########################################################
                    
                    if k <= 1:
                        pk=ones((5)) / num_strategy
                        rk=np.linspace(0,1,num_strategy+1)
                        success_mem=zeros((num_strategy))
                        failure_mem=zeros((num_strategy))
                        sk=zeros((num_strategy))
                    else:
                        if update_flag:
                            total=success_mem + failure_mem
                            total[np.where(total == 0)[0]]=1
                            sk=(success_mem / total) + 0.01
                            pk=sk / np.sum(sk)
                            rk=np.array([0]  + list(np.cumsum(pk)))
                            #print('success=', success_mem)
                            #print('fail=', failure_mem)
                            
                            success_mem=zeros(num_strategy)
                            failure_mem=zeros(num_strategy)
                            update_flag=False
                            
                    probability=uniform()
                    
                    if probability >= rk[0] and probability < rk[1]:
                        #---------------
                        #--- classic PSO  
                        #---------------
                        strategy_k=0
                        delta[i,:]=c2_1[k] * uniform(size=self.dim) * (self.pbest_pos[i,:] - self.Positions[i,:]) + c2_2[k] * uniform(size=self.dim) * (gbest_pos_temp[i,:] - self.Positions[i,:])
                        self.vel[i,:]= w2[k] * self.vel[i,:] + delta[i,:]
                    elif probability >= rk[1] and probability < rk[2]:
                        #---------------
                        #---  FDR-PSO
                        #---------------
                        strategy_k=1
                        dis=abs(np.tile(self.pbest_pos[i,:],(self.num_g2,1)) - self.pbest_pos[:self.num_g2,:])
                        fiterr=np.tile(self.pbest_val[i],(self.num_g2)) - self.pbest_val[:self.num_g2]
                        fiterr=fiterr.reshape(-1,1) * ones((self.num_g2,self.dim))
                        fiterr=fiterr - (dis == zeros((self.num_g2,self.dim))) * fiterr
                        dis=dis + (dis == zeros((self.num_g2,self.dim)))
                        FDR=fiterr / dis
                        _ ,Fid=np.max(FDR,axis=0), np.argmax(FDR,axis=0)
                        
                        Pnd=np.array([self.pbest_pos[Fid[dimcnt],dimcnt] for dimcnt in range(self.dim)])
                        
                        delta[i,:]=fii[0] * uniform(size=self.dim) * (self.pbest_pos[i,:] - self.Positions[i,:]) \
                                   + fii[1] * uniform(size=self.dim) * (gbest_pos_temp[i,:] - self.Positions[i,:])  \
                                   + fii[2] * uniform(size=self.dim) * (Pnd - self.Positions[i,:])
                        self.vel[i,:]=w2[k] * self.vel[i,:] + delta[i,:]
        
                    elif probability >= rk[2] and probability < rk[3]:
                        #---------------
                        #---  HPSO
                        #---------------
                        strategy_k=2
                        self.vel[i,:]= c3_1[k] * uniform(size=self.dim) * (self.pbest_pos[i,:] - self.Positions[i,:]) + c3_2[k] * uniform(size=self.dim) * (gbest_pos_temp[i,:] - self.Positions[i,:])
                        for d in range(self.dim):
                            if self.vel[i,d] == 0:
                                if uniform() < 0.5:
                                    self.vel[i,d]=uniform() * re_init_vel[k,d]
                                else:
                                    self.vel[i,d]=-uniform() * re_init_vel[k,d]
                                    
                            self.vel[i,d]=np.sign(self.vel[i,d]) * np.minimum(abs(self.vel[i,d]),self.ub[d])
                            
                    elif probability >= rk[3] and probability < rk[4]:
                        #---------------
                        #---  LIPS
                        #---------------
                        strategy_k=3
                        EU_dist=np.array([np.linalg.norm(self.Positions[i,:] - ind) for ind in self.pbest_pos])   #calculate Euclidean distance
                        EU_dist[i]=np.max(EU_dist)
                        _ ,min_index=np.sort(EU_dist), np.argsort(EU_dist)
                        fi=(4.1 / nsize) * uniform(size=(nsize,self.dim))
                        FIP=sum(multiply(fi,self.pbest_pos[min_index[0:nsize],:])) / sum(fi)
                        delta[i,:]=sum(fi) * (FIP - self.Positions[i,:])
                        self.vel[i,:]=0.7298 * (self.vel[i,:] + delta[i,:])
                                    #
                    elif probability >= rk[4] and probability < rk[5]:
                        #---------------
                        #---  CLPSO
                        #---------------
                        strategy_k=4
                        self.UpdateParticles(a=i, b=i+1, friend_num=self.num_g1, check_slope=True)
                        delta[i,:]=c4_1[k] * uniform(size=self.dim) * (self.fri_best_pos[i,:] - self.Positions[i,:]) \
                                   + c4_2[k] * uniform(size=self.dim) * (gbest_pos_temp[i,:] - self.Positions[i,:])
                        self.vel[i,:]=w2[k] * self.vel[i,:] + delta[i,:]
        
                    # forall
                    self.vel[i,:]=(multiply((self.vel[i,:] < v_min[i,:]),v_min[i,:]))  \
                            + (multiply((self.vel[i,:] > v_max[i,:]),v_max[i,:])) \
                            + (multiply((np.logical_and((self.vel[i,:] < v_max[i,:]),(self.vel[i,:] > v_min[i,:]))),self.vel[i,:])) 
                    self.Positions[i,:]=self.Positions[i,:] + self.vel[i,:]
                    #for discrete mutation
                    self.Positions[i, :] = self.ensure_bounds(self.Positions[i, :])
                    self.Positions[i, :] = self.ensure_discrete(self.Positions[i, :])
                    
                    strat.append(strategy_k)
                    
                #evaluate all individuals
                result2=self.eval_particles(pop=self.Positions[self.num_g1:,:])
                result = result + result2  #append second group fitness 
                
                idx=0
                for i in range(self.num_g1, self.num_g):
                    if result[i] < self.pbest_val[i]:
                        self.pbest_pos[i,:]=self.Positions[i,:]
                        self.pbest_val[i]=result[i]
                        success_mem[strat[idx]]=success_mem[strat[idx]] + 1
                    else:
                        failure_mem[strat[idx]]=failure_mem[strat[idx]] + 1
                        
                    if strat[idx] == 4 and result[i] < self.pbest_val[i]:
                        self.obj_func_slope[i]=0
                    else:
                        self.obj_func_slope[i]=self.obj_func_slope[i] + 1
        
                    if self.pbest_val[i] < self.gbest_val:
                        self.gbest_pos=self.pbest_pos[i,:]
                        self.gbest_val=self.pbest_val[i]
                    
                    idx+=1
                
                if (k + 1) % LP == 0:
                    update_flag=True
                    
                ###########################################################
                #END OF IDEAL SECTION
                ###########################################################
        
            #----------------------------------------------
            # Phase 2: Exploitation
            #----------------------------------------------
            else:
                #print('--- Gen %i: EPSO Exploitation Phase'%k)
                
                for i in range(self.num_g):
                    
                    ###########################################################
                    #DONT FORGET TO CHANGE IN Exploration PHASE.
                    #every line below this is a copy of what is in exploration
                    ###########################################################
                    
                    if k <= 1:
                        pk=ones((5)) / num_strategy
                        rk=np.linspace(0,1,num_strategy+1)
                        success_mem=zeros((num_strategy))
                        failure_mem=zeros((num_strategy))
                        sk=zeros((num_strategy))
                    else:
                        if update_flag:
                            total=success_mem + failure_mem
                            total[np.where(total == 0)[0]]=1
                            sk=(success_mem / total) + 0.01
                            pk=sk / np.sum(sk)
                            rk=np.array([0]  + list(np.cumsum(pk)))
                            #print('success=', success_mem)
                            #print('fail=', failure_mem)
                            #print('rk=', rk)
                            
                            success_mem=zeros(num_strategy)
                            failure_mem=zeros(num_strategy)
                            update_flag=False
        
                    probability=uniform()
        
                    if probability >= rk[0] and probability < rk[1]:
                        #---------------
                        #--- classic PSO  
                        #---------------
                        strategy_k=0
                        delta[i,:]=c2_1[k] * uniform(size=self.dim) * (self.pbest_pos[i,:] - self.Positions[i,:]) + c2_2[k] * uniform(size=self.dim) * (gbest_pos_temp[i,:] - self.Positions[i,:])
                        self.vel[i,:]= w2[k] * self.vel[i,:] + delta[i,:]
                        
                    elif probability >= rk[1] and probability < rk[2]:
                        #---------------
                        #---  FDR-PSO
                        #---------------
                        strategy_k=1
                        dis=abs(np.tile(self.pbest_pos[i,:],(self.num_g2,1)) - self.pbest_pos[:self.num_g2,:])
                        fiterr=np.tile(self.pbest_val[i],(self.num_g2)) - self.pbest_val[:self.num_g2]
                        fiterr=fiterr.reshape(-1,1) * ones((self.num_g2,self.dim))
                        fiterr=fiterr - (dis == zeros((self.num_g2,self.dim))) * fiterr
                        dis=dis + (dis == zeros((self.num_g2,self.dim)))
                        FDR=fiterr / dis
                        _ ,Fid=np.max(FDR,axis=0), np.argmax(FDR,axis=0)
                        
                        Pnd=np.array([self.pbest_pos[Fid[dimcnt],dimcnt] for dimcnt in range(self.dim)])
                        
                        delta[i,:]=fii[0] * uniform(size=self.dim) * (self.pbest_pos[i,:] - self.Positions[i,:]) \
                                   + fii[1] * uniform(size=self.dim) * (gbest_pos_temp[i,:] - self.Positions[i,:])  \
                                   + fii[2] * uniform(size=self.dim) * (Pnd - self.Positions[i,:])
                        self.vel[i,:]=w2[k] * self.vel[i,:] + delta[i,:]
                        
                    elif probability >= rk[2] and probability < rk[3]:
                        #---------------
                        #---  HPSO
                        #---------------
                        strategy_k=2
                        self.vel[i,:]= c3_1[k] * uniform(size=self.dim) * (self.pbest_pos[i,:] - self.Positions[i,:]) + c3_2[k] * uniform(size=self.dim) * (gbest_pos_temp[i,:] - self.Positions[i,:])
                        for d in range(self.dim):
                            if self.vel[i,d] == 0:
                                if uniform() < 0.5:
                                    self.vel[i,d]=uniform() * re_init_vel[k,d]
                                else:
                                    self.vel[i,d]=-uniform() * re_init_vel[k,d]
                                    
                            self.vel[i,d]=np.sign(self.vel[i,d]) * np.minimum(abs(self.vel[i,d]),self.ub[d])
                            
                    elif probability >= rk[3] and probability < rk[4]:
                        #---------------
                        #---  LIPS
                        #---------------
                        strategy_k=3
                        EU_dist=np.array([np.linalg.norm(self.Positions[i,:] - ind) for ind in self.pbest_pos])   #calculate Euclidean distance
                        EU_dist[i]=np.max(EU_dist)
                        _ ,min_index=np.sort(EU_dist), np.argsort(EU_dist)
                        fi=(4.1 / nsize) * uniform(size=(nsize,self.dim))
                        FIP=sum(multiply(fi,self.pbest_pos[min_index[0:nsize],:])) / sum(fi)
                        delta[i,:]=sum(fi) * (FIP - self.Positions[i,:])
                        self.vel[i,:]=0.7298 * (self.vel[i,:] + delta[i,:])
                        
                    elif probability >= rk[4] and probability < rk[5]:
                        #---------------
                        #---  CLPSO
                        #---------------
                        strategy_k=4
                        self.UpdateParticles(a=i, b=i+1, friend_num=self.num_g1, check_slope=True)
                        delta[i,:]=c4_1[k] * uniform(size=self.dim) * (self.fri_best_pos[i,:] - self.Positions[i,:]) \
                                   + c4_2[k] * uniform(size=self.dim) * (gbest_pos_temp[i,:] - self.Positions[i,:])
                        self.vel[i,:]=w2[k] * self.vel[i,:] + delta[i,:]
                        
                    # forall
                    self.vel[i,:]=(multiply((self.vel[i,:] < v_min[i,:]),v_min[i,:]))  \
                            + (multiply((self.vel[i,:] > v_max[i,:]),v_max[i,:])) \
                            + (multiply((np.logical_and((self.vel[i,:] < v_max[i,:]),(self.vel[i,:] > v_min[i,:]))),self.vel[i,:])) 
                    self.Positions[i,:]=self.Positions[i,:] + self.vel[i,:]
                    #for discrete mutation
                    self.Positions[i, :] = self.ensure_bounds(self.Positions[i, :])
                    self.Positions[i, :] = self.ensure_discrete(self.Positions[i, :])
                    strat.append(strategy_k)
                    
                #evaluate all individuals
                result=self.eval_particles(pop=self.Positions)                
                
                #evaluate all individuals
                idx=0
                for i in range(self.num_g):
                    if result[i] < self.pbest_val[i]:
                        self.pbest_pos[i,:]=self.Positions[i,:]
                        self.pbest_val[i]=result[i]
                        success_mem[strat[idx]]=success_mem[strat[idx]] + 1
                    else:
                        failure_mem[strat[idx]]=failure_mem[strat[idx]] + 1
                        
                    if strat[idx] == 4 and result[i] < self.pbest_val[i]:
                        self.obj_func_slope[i]=0
                    else:
                        self.obj_func_slope[i]=self.obj_func_slope[i] + 1
        
                    if self.pbest_val[i] < self.gbest_val:
                        self.gbest_pos=self.pbest_pos[i,:]
                        self.gbest_val=self.pbest_val[i]
                    
                    idx+=1
                
                if (k + 1) % LP == 0:
                    update_flag=True   
                    
                ###########################################################
                #END OF IDEAL SECTION
                ###########################################################
                
            success_hist[k,:] = pk
                
                
            #--mir
            self.best_position=self.gbest_pos.copy()
            self.best_fitness=self.gbest_val.copy()
            
            if self.mode=='max':
                self.fitness_best_correct=-self.best_fitness
                self.local_fitness=-np.min(result)
            else:
                self.fitness_best_correct=self.best_fitness
                self.local_fitness=np.min(result)

            self.last_pop=self.Positions.copy()
            self.last_fit=np.array(result).copy()            
            self.history['local_fitness'].append(self.local_fitness)
            self.history['global_fitness'].append(self.fitness_best_correct)
            
            skr=np.round(pk,2)
            success_rate={'PSO': skr[0], 'FDR-PSO': skr[1], 'HPSO':skr[2], 'LIPS':skr[3], 'CLPSO':skr[4]}
            
            # Print statistics
            if self.verbose:
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('EPSO step {}/{}, npar = g1+g2 ={}, Ncores={}'.format((k+1)*self.num_g, ngen*self.num_g, self.num_g, self.ncores))
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')
                print('Best Particle Fitness:', np.round(self.fitness_best_correct,6))
                if self.grid_flag:
                    self.particle_decoded = decode_discrete_to_grid(self.best_position, self.orig_bounds, self.bounds_map)
                    print('Best Particle Position:', self.particle_decoded)
                else:
                    print('Best Particle Position:', self.best_position)
                print('Success Rate:', success_rate)
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^')

        #mir-grid
        if self.grid_flag:
            self.epso_correct = decode_discrete_to_grid(self.best_position, self.orig_bounds, self.bounds_map)
        else:
            self.epso_correct = self.best_position                

        if self.mode=='max':
            self.last_fit=-self.last_fit
        
        #--mir return the last population for restart calculations
        if self.grid_flag:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=True, 
                                                     bounds=self.orig_bounds, bounds_map=self.bounds_map)
        else:
            self.history['last_pop'] = get_population(self.last_pop, fits=self.last_fit, grid_flag=False)
            
        success_hist=pd.DataFrame(success_hist, columns=['PSO', 'FDR-PSO', 'HPSO', 'LIPS', 'CLPSO'])
        self.history['success_rate'] = success_hist
        
        if self.verbose:
            print('------------------------ EPSO Summary --------------------------')
            print('Best fitness (y) found:', self.fitness_best_correct)
            print('Best individual (x) found:', self.epso_correct)
            print('--------------------------------------------------------------')  
            
        return self.epso_correct, self.fitness_best_correct, self.history

